#!/usr/bin/python3
"""
  For Production deployments, see: part_check instead of this file (file_check)  

  As check re-reads the entire file, re-calculates from scratch, which is very inefficient 
  compared to part_check which takes care of checksum calc done as the files are downloaded.
  NOTE: it also deletes the downloaded file after checking!

Usage:

   callback work.check

STATUS:
  This is more for debugging, internal testing, and reference.
  20171212: not sure if it has been fixed after the local->new transition.

"""
import os, stat, time
from hashlib import md5
import logging

import sarracenia
from sarracenia.flowcb import FlowCB

logger = logging.getLogger(__name__)

class Check(FlowCB):

    def reset_metrics(self):
        self.checksum_mismatches=0
        self.size_mismatches=0
        self.ok=0
        self.bad_content=0

    def on_start(self):
        logger.info("")
        self.reset_metrics()

    def content_check(self,local_file) -> bool:
        """
           return True if content is ok, false otherwise.
        """
        logger.info("no content check implemented")
        return True

    def after_work(self, worklist):

        for msg in worklist.ok:
    
            local_file = os.path.join( msg['new_dir'], msg['new_file'] )
            logger.info("start local file %s " % local_file )

            if 'fileOp' in msg:
                logger.warning("ignore unordinary files fileOps")
                if 'directory' in msg['fileOp']:
                    os.unlink(local_file)
                else:
                    os.unlink(local_file)
                continue

            logger.info("identity     %s " % msg['identity'] )
            logger.info("filesize   %s " % msg['size'])
    
            lstat = os.stat(local_file)
            fsiz = lstat[stat.ST_SIZE]
    
            if fsiz != msg['size']:
                logger.error( "filesize differ (corrupted ?)  lf %d  msg %d" % (fsiz, msg['size']) )
                self.size_mismatches+=1
    
            self.o.post_baseUrl = msg['baseUrl']
            self.o.identity_method = msg['identity']['method']
            downloaded_msg = sarracenia.Message.fromFileData( local_file, self.o, lstat ) 
         
      
            if downloaded_msg['identity'] != msg['identity']:
                logger.error(
                    "checksum differ (corrupted ?)  lf %s  msg %s" %
                    (downloaded_msg['identity'], msg['identity']))
                self.checksum_mismatches+=1

            if self.content_check(local_file):
                os.unlink(local_file)
                logger.info( f"checked out ok {local_file}" )
                self.ok+=1
            else:
                self.bad_content+=1

    def metricsReport(self) -> dict:
        return { 'checked_ok': self.ok, 
                 'checked_size_mismatches': self.size_mismatches, 
                 'checked_checksum_mismatches': self.checksum_mismatches,
                 'checked_bad_content': self.bad_content
                 }

    def on_housekeeping(self) -> None:
        self.reset_metrics()



#
# This file is part of sarracenia.
# The sarracenia suite is Free and is proudly provided by the Government of Canada
# Copyright (C) Her Majesty The Queen in Right of Canada, Environment Canada, 2008-2015
#
# Sarracenia repository: https://github.com/MetPX/sarracenia
# Documentation: https://github.com/MetPX/sarracenia
#
########################################################################
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; version 2 of the License.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
#
#

"""
  Extra Feature Scanning and Enablement.

  checking for extra dependencies, not "hard" dependencies ("requires")
  listed as features in setup.py and omitted entirely from debian packaging.
  this allows installation with fewer dependencies ahead of time, and then
  provide some messaging to users when they "need" an optional dependency.
 
  optional features can be enabled using pip install metpx-sr3[extra]
  where extra is one of the features listed below. Alternatively,
  one can just install the modules that are needed and the functionality
  will be enabled after a component restart.
  
  amqp - ability to communicate with AMQP (rabbitmq) brokers
  mqtt - ability to communicate with MQTT brokers
  filetypes - ability to
  ftppoll - ability to poll FTP servers
  vip  - enable vip (Virtual IP) settings to implement singleton processing
         for high availability support.
  watch - monitor files or directories for changes.

"""

import logging
import sys

logger = logging.getLogger(__name__)


features = { 
    'amqp' : { 'modules_needed': [ 'amqp' ], 'present': False, 
            'lament' : 'cannot connect to rabbitmq brokers', 
            'rejoice' : 'can connect to rabbitmq brokers' },
   'appdirs' : { 'modules_needed': [ 'appdirs' ], 'present': False, 
           'lament' : 'assume linux file placement under home dir', 
           'rejoice': 'place configuration and state files appropriately for platform (windows/mac/linux)', },
   'filetypes' : { 'modules_needed': ['magic'], 'present': False, 
           'lament': '(pip package python-magic, on windows python-magic-bin) will not be able to set content headers' ,
           'rejoice': 'able to set content headers' },
   'ftppoll' : { 'modules_needed': ['dateparser', 'pytz'], 'present': False, 
       'lament' : 'not able to poll with ftp' ,
       'rejoice' : 'able to poll with ftp' },
   'humanize' : { 'modules_needed': ['humanize', 'humanfriendly' ], 'present': False, 
           'lament': 'humans will have to read larger, uglier numbers',
           'rejoice': 'humans numbers that are easier to read.' },
   'mqtt' : { 'modules_needed': ['paho.mqtt.client'], 'present': False, 
           'lament': 'cannot connect to mqtt brokers' ,
           'rejoice': 'can connect to mqtt brokers' },
   'process' : { 'modules_needed': ['psutil'], 'present': False,
        'lament': 'cannot monitor running processes, sr3 CLI basically does not work.',
        'rejoice': 'can monitor, start, stop processes:  Sr3 CLI should basically work' },
   'redis' : { 'modules_needed': [ 'redis', 'redis_lock' ],
        'lament': 'cannot use redis implementations of retry and nodupe',
        'rejoice': 'can use redis implementations of retry and nodupe'
        },
   'retry': { 'modules_needed': ['jsonpickle'], 'present': False,
           'lament': 'unable to use local queues on the side (disk or redis) to retry messages later',
           'rejoice': 'can write messages to local queues to retry failed publishes/sends/downloads'},
   'sftp' : { 'modules_needed': [ 'paramiko' ],
        'lament': 'cannot use or access sftp/ssh based services',
        'rejoice': 'can use sftp or ssh based services'
        },
   'vip'  : { 'modules_needed': ['netifaces'] , 'present': False, 
           'lament': 'will not be able to use the vip option for high availability clustering' ,
           'rejoice': 'able to use the vip option for high availability clustering' },
   'watch'  : { 'modules_needed': ['watchdog'] , 'present': False, 
           'lament': 'cannot watch directories' ,
           'rejoice': 'watch directories' },
   'xattr'  : { 'modules_needed': ['xattr'] , 'present': False, 
           'lament': 'unable to store additional file metadata in extended attributes' ,
           'rejoice': 'will store file metadata in extended attributes' }
}

if sys.platform == 'win32':
    features['ctypes'] = { 'modules_needed':['ctypes'], 'present':False, 
           'lament': 'unable to store additional file metadata in extended attributes' ,
           'rejoice': 'will store file metadata in extended attributes' }
   
for x in features:
   
   features[x]['present']=True
   for y in  features[x]['modules_needed']:
       try:
           # used to use importlib, but that only tested if the module itself was installed.
           # https://github.com/MetPX/sarracenia/issues/753
           # It turns out that import can fail because of transitive deps, a package is installed,
           # but something that it depends on is missing, so it's clearer to just try an import.

           exec( f"import {y}" )
           logger.debug( f'found feature {y}, which should help to enable {x}')
       except:
           logger.debug( f"extra feature {x} needs missing module {y}. Disabled" ) 
           features[x]['present']=False




# This file is part of sarracenia.
# The sarracenia suite is Free and is proudly provided by the Government of Canada
# Copyright (C) Her Majesty The Queen in Right of Canada, Environment Canada, 2008-2015
#
# more info: https://github.com/MetPX/sarracenia
#
# Code originally contributed by:
#  Michel Grenier - Shared Services Canada
#  first shot     : Wed Jan 10 16:06:16 UTC 2018
#  re-factored beyond recognition by PSilva 2021. Don't blame Michel
#

from _codecs import decode, encode

import jsonpickle, sarracenia, sys, time
import redis, redis_lock
import re

import logging

# class sarra/retry

logger = logging.getLogger(__name__)


class RedisQueue():
    """
    Process Persistent Queue...

    Persist messages to a Redis List so that processing can be attempted again later.
    So continuous operation, with an occasional housekeeping cycle to resolve them

    retry_ttl how long 

    self.retry_cache 
    * a dictionary indexed by some sort of key to prevent duplicate
        messages being stored in it.

    If <queueName> isn't specified
    key_name = 'sr3.retry_queue.<name>.<self.o.component>.<self.o.config>
    Otherise it's
    key_name = 'sr3.retry_queue.<name>.<self.o.component>.<self.o.queueName>

    It also gets various suffixes:

    .new -- messages added to the retry are pushed to this list.
    .hk  -- temporary list used during on_housekeeping.

    whenever a message is added to the retry_cache, it is appended to a 
    cumulative list of entries to add to the retry list.  

    every housekeeping interval, the two lists are consolidated.

    note that the *ack_id* of messages retreived from the retry list, is 
    removed. Files must be acked around the time they are placed on the 
    retry_list, as reception from the source should have already been acknowledged.
    """

    # ----------- magic Methods ------------
    def __init__(self, options, name):

        logger.debug(" %s __init__" % (name))

        self.o = options

        self.name = name

        if not hasattr(self.o, 'retry_ttl'):
            self.o.retry_ttl = None

        #logging.basicConfig(format=self.o.logFormat,
        #                    level=getattr(logging, self.o.logLevel.upper()))
        logger.setLevel(getattr(logging, self.o.logLevel.upper()))

        logger.debug('name=%s logLevel=%s' % (self.name, self.o.logLevel))

        if self.o.queueName == None: 
            self.key_name = 'sr3.retry_queue.' + name + '.' + self.o.component + '.' + self.o.config
        else:
            # #"autogenerated" queueName format taken from config.py
            # test_queueName = 'q_' + self.o.broker.url.username + '_' + self.o.component + '\.' + self.o.config
            # if hasattr(self.o, 'queue_suffix') and self.o.queue_suffix != None:
            #     test_queueName += '\.' + self.o.queue_suffix
            # test_queueName += '\.[0-9]{8}\.[0-9]{8}'
            # match = re.search(test_queueName ,self.o.queueName)
            # if match:
            #     logger.warn("Detected possible autogenerated queueName %s; should set it in config to avoid issues" % (self.o.queueName))
            
            self.key_name = 'sr3.retry_queue.' + name + '.' + self.o.component + '.' + self.o.queueName

        self.now = sarracenia.nowflt()

        # newer retries
        self.key_name_new = self.key_name + '.new'

        # working file at housekeeping
        self.key_name_hk = self.key_name + '.hk'

        # track last housekeeping to ensure that multiple instances don't try to do it more often than the defined interval
        self.key_name_lasthk = 'sr3.retry_queue.' + name + '.' + self.o.component + '.' + self.o.config + ".last_hk"

        self.o.add_option( 'redisqueue_serverurl', 'str')

        self.redis = redis.from_url(self.o.redisqueue_serverurl)

        self.redis.set(self.key_name_lasthk, self.now)

        # The lock will be set with a TTL of 'expire', but will auto renew at expire*2/3
        #  This should allow a crashed/failed/terminated instances' locks to clear out so others can do it
        self.redis_lock = redis_lock.Lock(self.redis, self.key_name_hk, expire=30, auto_renewal=True)

        # initialize ages and message counts


    def __len__(self) -> int:
        """Returns the total number of messages in the RedisQueue.

        Number of messages in the RedisQueue does not necessarily equal the number of messages available to ``get``.
        Messages in the .new list are counted, but can't be retrieved until
        :func:`~sarracenia.redisqueue.RedisQueue.on_housekeeping` has been run.

        Returns:
            int: number of messages in the RedisQueue.
        """
        return self.redis.llen(self.key_name) + self.redis.llen(self.key_name_new) 


    # ----------- Private Methods -----------
    def _in_cache(self, message) -> bool:
        """
        return whether the entry is message is in the cache or not.
        side effect: adds it.

        """
        urlstr = message['baseUrl'] + '/' + message['relPath']

        if 'noDupe' in message:
            sumstr = jsonpickle.encode(message['noDupe']['key'])
        elif 'fileOp' in message:
            sumstr = jsonpickle.encode(message['fileOp'])
        elif 'identity' in message:
            sumstr = jsonpickle.encode(message['identity'])
        elif 'pubTime' in message:
            sumstr = jsonpickle.encode(message['pubTime'])
        else:
            logger.info('no key found for message, cannot add')
            return False

        cache_key = urlstr + ' ' + sumstr

        if 'parts' in message:
            cache_key += ' ' + message['parts']

        if cache_key in self.retry_cache: return True
        self.retry_cache[cache_key] = True
        return False

    def _is_expired(self, message) -> bool:
        """
        return is the given message expired ?
        """
        # no expiry
        if self.o.retry_ttl is None: return False
        if self.o.retry_ttl <= 0: return False

        # compute message age
        msg_time = sarracenia.timestr2flt(message['pubTime'])
        msg_age = self.now - msg_time

        # expired ?
        return msg_age > self.o.retry_ttl

    def _needs_requeuing(self, message) -> bool:
        """
        return 
            * True if message is not expired, and not already in queue. 
            * False otherwise.   
        """
        if self._in_cache(message):
            logger.info("discarding duplicate message (in %s cache) %s" % (self.name, message))
            return False

        # log is info... it is good to log a retry message that expires
        if self._is_expired(message):
            logger.info("discarding expired message in (%s): %s" % (self.name, message))
            return False

        return True


    def _msgFromJSON(self, message):
        try:
            msg = jsonpickle.decode(message)
        except ValueError:
            logger.error("corrupted item in list: %s " % (message))
            logger.debug("Error information: ", exc_info=True)
            return None
        except TypeError:
            logger.error("wrong type item in list: %s " % (message))
            logger.debug("Error information: ", exc_info=True)
            return None

        return msg

    def _msgToJSON(self, message):
        return jsonpickle.encode(message)

    def _lpop(self, queue):

        raw_msg = self.redis.lpop(queue)
          
        if raw_msg == None:
            return None

        msg = self._msgFromJSON(raw_msg)
        logger.debug("lpop from list %s %s" % (queue, msg))

        return msg


    # ----------- Public Methods -----------
    def put(self, message_list):
        """
        add messages to the end of the queue.
        """

        for message in message_list:
            logger.debug("rpush to list %s %s" % (self.key_name_new, message))
            self.redis.rpush(self.key_name_new, self._msgToJSON(message))

    def cleanup(self):
        """
        remove statefiles.
        """
        logger.debug("Starting redis cleanup")
        self.redis_lock.reset()
        self.redis.delete(self.key_name_lasthk)
        self.redis.delete(self.key_name_new)
        self.redis.delete(self.key_name_hk)
        self.redis.delete(self.key_name)
        logger.debug("done redis cleanup")

    def close(self):
        """
        clean shutdown.
        """
        #self.redis_lock.reset()

        self.redis = None

    def get(self, maximum_messages_to_get=1):
        """
        qty number of messages to retrieve from the queue.
        """

        ml = []

        if self.redis_lock.locked():
            logger.debug("Can't pop because housekeeping lock is active")
            return ml

        count = 0
        while count < maximum_messages_to_get:

            message = self._lpop(self.key_name)
            if not message: break
                
            if self._is_expired(message):
                logger.warn("message expired %s" % (message))
                continue

            if 'ack_id' in message:
                del message['ack_id']
                message['_deleteOnPost'].remove('ack_id')

            ml.append(message)
            count += 1

        return ml

    def on_housekeeping(self):
        """
        read rest of queue_file (from current point of unretried ones.)
            - check if message is duplicate or expired.
            - write to .hk

        read .new file, 
            - check if message is duplicate or expired.
            - writing to .hk (housekeeping)

        remove .new

        rename housekeeping to queue for next period.
        """
        logger.info("%s on_housekeeping" % (self.name))

        if float(self.redis.get(self.key_name_lasthk)) + self.o.housekeeping > sarracenia.nowflt():
            logger.info("Housekeeping ran less than %ds ago; not running " % (self.o.housekeeping))
            return

        # A shared/distributed locking system is required when using Redis
        #  because only a single instance of a config should ever run the Housekeeping tasks
        if self.redis_lock.locked():
            logger.info("Another instance has lock on %s" % (self.key_name_hk))
            while self.redis_lock.locked():
                time.sleep(1)
            return

        self.redis_lock.acquire()
        logger.info("got redis_lock %s" % (self.key_name_hk))

        # finish retry before reshuffling all retries entries
        if self.redis.llen(self.key_name) > 0:
            logger.info("have not finished retry list; resuming retries from %s" % (self.key_name))
            return

        self.now = sarracenia.nowflt()
        self.retry_cache = {}
        N = 0

        # put this in try/except in case ctrl-c breaks something
        try:
            try:
                logger.debug('delete list: %s' % (self.key_name_hk))
                self.redis.delete(self.key_name_hk)
            except:
                pass

            i = 0

            logger.debug("%s has queue %s" % (self.key_name, bool(self.redis.llen(self.key_name))))

            # remaining of retry to housekeeping
            while True:

                message = self._lpop(self.key_name)
                if not message: break

                i = i + 1
                if not self._needs_requeuing(message): continue

                logger.debug("remaining of retry - rpush to %s %s" % (self.key_name_hk, message))
                self.redis.rpush(self.key_name_hk, self._msgToJSON(message))
                N = N + 1

            i = 0
            j = N

            # append new to housekeeping.
            while True:

                message = self._lpop(self.key_name_new)
                if not message: break

                i = i + 1
                #logger.debug("DEBUG message %s" % message)
                if not self._needs_requeuing(message): continue

                logger.debug("new to hk - rpush to %s %s" % (self.key_name_hk, message))
                self.redis.rpush(self.key_name_hk, self._msgToJSON(message))
                N = N + 1

            logger.debug("FIXME DEBUG took %d out of the %d retry" % (N - j, i))

        except Exception as Err:
            logger.error("something went wrong")
            logger.debug('Exception details: ', exc_info=True)

        # no more retry
        if N == 0:
            logger.info("No retry in list")
            try:
                logger.debug('no more retry - delete list: %s' % (self.key_name_hk))
                self.redis.delete(self.key_name_hk)
            except:
                pass

        # housekeeping file becomes new retry
        else:
            logger.info("Number of messages in retry list %d" % (N))

            try:
                logger.debug('rename list %s to %s' % (self.key_name_hk, self.key_name))
                self.redis.rename(self.key_name_hk, self.key_name)

            except Exception as Err:
                logger.error("Something went wrong with rename")
                logger.debug('Exception details: ', exc_info=True)


        self.redis.set(self.key_name_lasthk, self.now)
        self.redis_lock.release()
        logger.debug("released redis_lock")

        elapse = sarracenia.nowflt() - self.now
        logger.info("on_housekeeping elapse %f" % (elapse))



# This file is part of sarracenia.
# The sarracenia suite is Free and is proudly provided by the Government of Canada
# Copyright (C) His Majesty The King in Right of Canada, Environment Canada, 2008-2020
#

"""
Description:
    Sundew migration

    This is a sr3 plugin built to migrate the sender of ECCC's
    proprietary Alpha Manager(AM) socket protocol.

    For more information on the origins of AM and Sundew visit https://github.com/MetPX/Sundew/blob/main/doc/historical/Origins.rst

    Code overview:
        To start, when on_start is called, the socket is instiated and attempts to connect with a remote host with the IP and port # specified in the config file.
        Once the send method is called, the bulletin is wrapped with a 128 byte header which contains the bulletin length, thread number, timestamps, etc.
        The wrapped bulletin is then individually sent through the socket.
        If the send is successful, return True. If not, return False.

    Options:
        MaxBulLen (int): Specify the maximum length of the bulletin. If the bulletin is too big, reject.

    NOTE: AM cannot correct data corruption and AM cannot be stopped without data loss.
    Precautions should be taken during maintenance interventions.

Usage:
    flowcb sarracenia.flowcb.send.am.AM
    See sarracenia/examples/sender/am_send.conf for an example config file.

Author:
    André LeBlanc, ANL, Autumn 2022
"""

import logging, socket, struct, time, signal, sys, os
import urllib.parse
from sarracenia.flowcb import FlowCB

logger = logging.getLogger(__name__)

class AM(FlowCB):
    
    def __init__(self, options):
             
        super().__init__(options,logger)

        self.url = urllib.parse.urlparse(self.o.sendTo)

        self.threadnum = 255
        self.host = self.url.netloc.split(':')[0]
        self.port = int(self.url.netloc.split(':')[1])
        self.patternAM = '80sII4siIII20s'

        self.o.add_option('MaxBulLen', 'count', 32768)

        # Initialise socket
        ## Create a TCP/IP socket
        self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 2)

        # Add signal handler
        ## Override outer signal handler with a default one to exit correctly.
        signal.signal(signal.SIGTERM, signal.SIG_DFL)


    def wrapbulletin(self, sarra_msg): 

        logger.info("Commencing message wrap.")

        s = struct.Struct(self.patternAM)
        size = struct.calcsize('80s')

        msg_path = sarra_msg['new_relPath']
        msg_file = open(os.sep + msg_path, 'rb')
        data = msg_file.read()
        msg_file.close()

        # Construct msg header
        strdata = data.decode('iso-8859-1')
        header = strdata[0:size]

        # Step out of the function if the bulletin size is too big
        if len(strdata) > self.o.MaxBulLen:
            raise Exception(f"Bulletin length too long. Bulletin limit length: {self.o.MaxBulLen}. Latest bulletin length: {len(strdata)}. Path to bulletin: {msg_path}")

        ## Attach rest of header with NULLs (if not long enough)
        nulheader = ['\0' for _ in range(size)]
        nulheaderstr = ''.join(nulheader)
        header = header + nulheaderstr[len(header):]

        ## Perform bite swaps AND init miscellaneous header parameters
        length = socket.htonl(len(strdata.lstrip('\n')))

        firsttime = socket.htonl(int(time.time()))
        timestamp = socket.htonl(int(time.time()))
        threadnum = chr(0) + chr(self.threadnum) + chr(0) + chr(0)
        future = '\0'
        start, src_inet, dst_inet = (0, 0, 0)

        # Wrap bulletin
        ## Replace first line feed with NULL
        packedheader = s.pack(header.replace('\n','\x00',1).encode('iso-8859-1'), src_inet, dst_inet, threadnum.encode('iso-8859-1'), start
                                   , length, firsttime, timestamp, future.encode('iso-8859-1'))
        
        msg = packedheader + data

        logger.debug("Message has been packed.")
        # llogger.debug(f"Message contents: {msg}")
        
        return msg

    def on_stop(self):
        self.s.close()

    def reEstablishConnection(self):
        # Use exponential backoff to try and connect or reconnect to remote host
        
        if self.host == 'None':
            raise Exception("No remote host specified.")

        logger.info("Trying to connect to remote host %s and port %d" % (str(self.host) , self.port))

        backoff_range = 1
        while True:
            try:
                # Initialise socket
                ## Create a TCP/IP socket
                self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                self.s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 2)

                time.sleep(1)
                self.s.connect((socket.gethostbyname(self.host), self.port))
                break
                
            except socket.error as e:
                logger.debug("Error msg: %s" % str(e.args))
                logger.error("Trying to establish connection in %d seconds" % (2**backoff_range))
                self.s.close()
                time.sleep(2**backoff_range)
                if backoff_range < 6:
                    backoff_range += 1


    def send(self, bulletin):
        try:
            self.packed_bulletin = self.wrapbulletin(bulletin)

            while True:
                try:
                    bytesSent = self.s.send(self.packed_bulletin)

                    # Check if went okay
                    return bytesSent == len(self.packed_bulletin)
                    
                except socket.error as e:
                    logger.debug("Bulletin not sent. Error message: %s",str(e.args))
                    logger.error("Connection interrupted. Attempting to reconnect")
                    self.s.close()
                    self.reEstablishConnection()

        except Exception as e:
            raise Exception("Generalized error handler. Error message: %s", str(e.args))



"""
Plugin posthourtree.py:
    When posting a file, insert an hourly directory into the delivery path hierarchy.

Example:
    input A/B/c.gif  --> output A/B/<hour>/c.gif

Usage:
    callback accept.posthourtree

"""
import logging
import sys, os, os.path, time, stat
from sarracenia.flowcb import FlowCB

logger = logging.getLogger(__name__)


class Posthourtree(FlowCB):
    def __init__(self, options):
        super().__init__(options, logger)

    def after_accept(self, worklist):
        for message in worklist.incoming:
            datestr = time.strftime('%H', time.gmtime())  # pick the hour
            # insert the hour into the rename header of the message to be posted.
            message['new_dir'] += '/' + datestr 
            logger.info(  f"post_hour_tree: new_dir: {message['new_dir']}" )



"""
Plugin dateappend.py:
    This is an  example of the usage of after_accept script
    to rename the product when it is downloaded
    It adds a '_datetime' to the name of the product

    unlikely to be useful to others, except as example.
    note that, in the case of retries, the date will be appended multiple times.
    
    This is for diagnosing problems with data pumps, identifying feed failures and retries.

Example:
    takes input name  : ./SK/s0000684_f.xml
    add datetimestamp : ./SK/s0000684_f.xml_20221022080652
    retries: /SK/s0000684_f.xml_20221022080652_20221022081403, ./SK/s0000684_f.xml_20221022080652_20221022081403_20221022081508

Usage:
    callback accept.dateappend

    if installing where the plugin is not from the package itself, invocation is:

    flowcallback dateappend.Dateappend


"""
import logging
import sys, os, os.path, time, stat
from sarracenia.flowcb import FlowCB

logger = logging.getLogger(__name__)

class Dateappend(FlowCB):
    def __init__(self, options):
        super().__init__(options,logger)

    def after_accept(self, worklist):
        for message in worklist.incoming:
            datestr = time.strftime('_%Y%m%d%H%M%S', time.localtime())
            message['new_file'] += datestr



#!/usr/bin/python3

"""
Plugin to poll NASA using the CMR "Common Metadata Repository" API
=====================================================================================

What is CMR?

    NASA's Common Metadata Repository (CMR) is a high-performance, high-quality, 
    continuously evolving metadata system that catalogs all data and service metadata 
    records for NASA's Earth Observing System Data and Information System (EOSDIS) 
    system and will be the authoritative management system for all EOSDIS metadata. 
    These metadata records are registered, modified, discovered, and accessed through
    programmatic interfaces leveraging standard protocols and APIs.
    
    -- https://www.earthdata.nasa.gov/eosdis/science-system-description/eosdis-components/cmr

.. IMPORTANT::
    Most URLs posted by this plugin will require an Earthdata account to download the data.
    https://urs.earthdata.nasa.gov/  The corresponding download configuration (e.g. a sarra or
    subscribe config) will need to use the ``auth_NASA_Earthdata`` after_accept plugin.

This code is based on https://github.com/podaac/tutorials/blob/master/notebooks/opendap/MUR-OPeNDAP.ipynb


  
Configurable Options:
----------------------

``collectionConceptId`` Collection Concept IDs:
^^^^^^^^^^^^^^^^^^^^^^^^

    Determine where the data comes from. You can find the CCID from a page like
    this one: https://podaac.jpl.nasa.gov/dataset/VIIRS_NPP-STAR-L3U-v2.80

    You can define multiple ``collectionConceptId`` s in your config file.

``dap_urlExtension`` URL extension:
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    **Ignored when** ``dataSource`` **is NOT opendap**.  

    From OpenDAP:

    Here is a list of the six URL extensions that are be recognized by all DAP servers:
        - dds - The DAP2 syntactic (structural) metadata response.
        - dds - The DAP2 semantic metadata response.
        - dods - The DAP2 data response.
        - info - The DAP2 HTML dataset information page.
        - html - The DAP2 HTML data request form.
        - ascii - The DAP2 ASCII data response.

    In addition Hyrax and other new servers support the following DAP4 URL extensions:
        - dmr - The DAP4 dataset metadata response.
        - dap - The DAP4 data response.
        - dsr - The DAP4 dataset services response.
        - ddx - The DAP3.2 dataset metadata response.

``dap_fileType`` File type:
^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    **Ignored when** ``dataSource`` **is NOT opendap**.  The PO.DAAC and other URLs will already 
    include a file extension.

    You can see the file types if you browse to the raw "post_urls" URLs, for example
    https://opendap.earthdata.nasa.gov/providers/POCLOUD/collections/GHRSST%20Level%204%20MUR%20Global%20Foundation%20Sea%20Surface%20Temperature%20Analysis%20(v4.1)/granules/20190101090000-JPL-L4_GHRSST-SSTfnd-MUR-GLOB-v02.0-fv04.1.dmr.html

    The supported options are:
        - nc - NetCDF-3
        - nc4 - NetCDF-4
        - csv - CSV 
        - dap4b - DAP4 binary (the plugin will interpret this and set the extension in the URL to nothing)

``timeNowMinus`` Temporal (Time) Range:
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    The API requires that a date time range is specified when polling. The plugin will always use the current 
    date/time as the "end time" and the "start time" is determined by the ``timeNowMinus`` duration option.

    This is a duration option. It will accept a floating point number, or a floating point number suffixed with
    a unit. Examples:
        - Bare number: ``timeNowMinus 3600`` --> 3600 seconds
        - ``s`` Seconds: ``timeNowMinus 3600`` --> 3600 seconds
        - ``m`` Minutes: ``timeNowMinus 120m`` --> 120 minutes
        - ``h`` Hours: ``timeNowMinus 24h`` --> 24 hours
        - ``d`` Days: ``timeNowMinus 5d`` --> 5 days
        - ``w`` Weeks: ``timeNowMinus 1w`` --> 1 week

    For example, if the poll is running at 2023-04-26 12:35:00Z and ``timeNowMinus`` is ``1d``, 
    the temporal range to poll would be: ::

        start,end
        2023-04-25T12:35:00Z,2023-04-26T12:35:00Z

Configurable data source:
^^^^^^^^^^^^^^^^^^^^^^^^^^

    The API returns multiple "related URLs" for each item. For the PO.DAAC data that this plugin was originally 
    written to acquire, there are both PO.DAAC sources (archive.podaac.earthdata.nasa.gov) and OPeNDAP.

    When ``dataSource`` in the poll config is set to ``podaac``, these pre-set parameters are used to find the URL:
        - Type: "GET DATA" (the API also provides "EXTENDED METADATA" URLs that can be used to get the md5sum)
        - Description: should contain "Download"
        - URL: should contain podaac

        For PO.DAAC, the plugin will also attempt to get and post the file's md5sum in the message.

    When ``dataSource`` is set to ``opendap``, the pre-set parameters are:
        - Description: contains "OPeNDAP"

        For OPeNDAP, you must set the ``dap_urlExtension`` and ``dap_fileType`` options in the poll config.
    
    When ``dataSource`` is ``other``, the parameters must be set in the config file:
        - Type: must be equal to ``relatedUrl_type`` option (required).
        - Description: contains ``relatedUrl_descriptionContains`` option, if defined.
        - URL: contains ``relatedUrl_urlContains`` option, if defined.

        ``relatedUrl_type`` must be set in the poll config. The other ``related_url...`` settings are optional.


*Potential Room For Improvement*: 
----------------------------------

    - To allow multiple dap_urlExtension and dap_fileTypes in one config, maybe post URLs with all items in the
      self.o.dap_urlExtension and self.o.dap_fileType lists, and then use accept and reject to filter...?
    - Also convert relatedUrl_type to list to support multiple types.

How to set up your poll config:
--------------------------------
 
    Use ``callback poll.poll_NASA_CMR``, and read about the config options above.
    
    For examples, see https://github.com/MetPX/sarracenia/tree/main/sarracenia/examples/poll files named ``*nasa_cmr*.conf``. 

Change log:
-----------

    - 2023-10-10: finished porting poll code, now the poll is in a working state.
    - 2023-06-29: ported to sr3, split into poll_NASA_CMR (incomplete) and download plugin.
    - 2023-06-14: added ``other`` dataSource. Renamed to nasa_cmr from nasa_opendap.
"""

import sarracenia
import datetime
import logging
import requests
from urllib.parse import urlparse

logger = logging.getLogger(__name__)

class Poll_nasa_cmr(sarracenia.flowcb.FlowCB):
    def __init__(self, options):
        super().__init__(options, logger)
        
        # Allow setting a logLevel *only* for this plugin in the config file:
        # set poll.poll_NASA_CMR.logLevel debug
        if hasattr(self.o, 'logLevel'):
            logger.setLevel(self.o.logLevel.upper())

        logger.debug("plugin: Poll_NASA_CMR __init__")

        self.o.add_option('collectionConceptId', kind='list') # required
        self.o.add_option('dataSource',          kind='str')  # required
        self.o.add_option('timeNowMinus',        kind='duration', default_value=3600.0) # 3600 sec = 1 hour
        self.o.add_option('pageSize',            kind='count', default_value=2000) 

        # Specific to dataSource=opendap
        self.o.add_option('dap_urlExtension', kind='str')
        self.o.add_option('dap_fileType',     kind='str')

        # Specific to dataSource=other
        self.o.add_option('relatedUrl_type',                kind='str') # Required, when dataSource=other
        self.o.add_option('relatedUrl_descriptionContains', kind='list') # Optional
        self.o.add_option('relatedUrl_urlContains',         kind='list') # Optional

        # Allowed settings for dataSource
        self._dataSource_options = ["podaac", "opendap", "other"]

        # pollUrl - if not specified by the user in the config, it should be this
        self._defaultPollUrl = "https://cmr.earthdata.nasa.gov/search/granules.umm_json"
        if not self.o.pollUrl:
            self.o.pollUrl = self._defaultPollUrl
        
        # The default of setting post_baseUrl = pollUrl is invalid for this plugin.
        # The baseUrl is never cmr.earthdata.nasa.gov... and there can be multiple different baseUrls, depending on
        # what URLs are returned by the API. The post_baseUrl setting will be set by the poll for each URL.
        self.o.post_baseUrl = None

        # Check required options
        if not self.o.dataSource or self.o.dataSource not in self._dataSource_options:
            logger.error(f"dataSource unknown: {self.o.dataSource}")
 
        if not self.o.collectionConceptId or len(self.o.collectionConceptId) < 1:
            logger.error(f"collectionConceptId is a required option. " +
                         f"{self.o.collectionConceptId} is invalid.")
        
        # dap_urlExtension and dap_fileType are required when using dataSource opendap
        if self.o.dataSource == "opendap":
            if not self.o.dap_urlExtension or len(self.o.dap_urlExtension) <= 0:
                logger.error(f"dap_urlExtension option must be set when using dataSource opendap.")
            if not self.o.dap_fileType or len(self.o.dap_fileType) <= 0:
                logger.error(f"dap_fileType option must be set when using dataSource opendap.")

        # relatedUrl_type is required when using dataSource other
        if self.o.dataSource == "other":
            if not self.o.relatedUrl_type or len(self.o.relatedUrl_type) <= 0:
                logger.error(f"relatedUrl_type option must be set when using dataSource other.")
        
        # end __init__

    def poll(self) -> list:
        """Poll NASA_CMR
        polling doesn't require authentication, only downloading does
        """
        logger.info("plugin: Poll_NASA_CMR starting poll")

        gathered_messages = []

        ### SETUP 

        # Build CMR URL to poll
        # Expecting https://cmr.earthdata.nasa.gov/search/granules.umm_json
        if self.o.pollUrl[-1] == '/':
            url_head = self.o.pollUrl[:-1]
        else:
            url_head = self.o.pollUrl
        if url_head != self._defaultPollUrl:
            logger.warning(f"The destination URL is {url_head}, is this right?")
        
        # No extension used if you want a DAP4 Binary File
        if self.o.dataSource == "opendap" and hasattr(self.o, "dap_fileType") and self.o.dap_fileType == "dap4b":
            dap_fileType = ""
        elif self.o.dataSource == "opendap": 
            dap_fileType = self.o.dap_fileType
        else:
            dap_fileType = ""
        
        # Check page size, 2000 is max
        pg_size = int(self.o.pageSize)
        if pg_size < 1 or pg_size > 2000:
            pageSize = "2000"
        else:
            pageSize = str(int(self.o.pageSize))

        # Figure out the date range
        # Format: start,end   2019-01-01T10:00:00Z,2019-02-01T00:00:00Z
        t_now = datetime.datetime.utcnow()
        t_range_end = t_now.strftime("%Y-%m-%dT%H:%M:%SZ")

        try:
            n_seconds = float(self.o.timeNowMinus)
        except:
            logger.error("invalid timeNowMinus option")
            n_seconds = 0

        t_now_minus = t_now - datetime.timedelta(seconds=n_seconds)
        t_range_start = t_now_minus.strftime("%Y-%m-%dT%H:%M:%SZ")
        
        temporal_range = t_range_start + "," + t_range_end
        logger.debug(f"Temporal range: {temporal_range} (timeNowMinus={self.o.timeNowMinus})")

        ### DO THE POLL

        # Example URL from GitHub:
        # https://cmr.earthdata.nasa.gov/search/granules.umm_json?collectionConceptId=C1625128926-GHRC_CLOUD&temporal=2019-01-01T10:00:00Z,2019-12-31T23:59:59Z

        for cci in self.o.collectionConceptId:
            url = url_head + "?collectionConceptId=" + cci
            # Date Time Range
            url += "&temporal=" + temporal_range
            url += "&pageSize=" + pageSize 
            logger.info(f"Polling URL: {url}")
            r = requests.get(url)
            response_body = r.json()

            # The response contains multiple URLs and other info.

            # Now can get the URLs
            for itm in response_body['items']:
                # There are different types of related URLs, including the actual files from PODAAC and OPeNDAP,
                # and metadata (e.g. md5 check sums)
                data_url = None
                md5_url = None
                for url in itm['umm']['RelatedUrls']:
                    if self.stop_requested:
                        logger.info("Received a request to stop. Aborting poll and returning an empty list.")
                        return []

                    logger.debug("inspecting potential URL: " + str(url))

                    # Use the kind of URL that was specified in the config
                    # Unknown
                    if self.o.dataSource not in self._dataSource_options:
                        logger.error(f"NASA_CMR plugin: NOT posting because dataSource unknown: {self.o.dataSource}")

                    # OPeNDAP
                    elif (self.o.dataSource == "opendap" and 'OPeNDAP' in url['Description'] ):
                        logger.debug(f"OPENDAP URL {url['URL']}")
                        # Add dap url extension and file type
                        data_url = url['URL'] + "." + self.o.dap_urlExtension + "." + dap_fileType

                    # PO.DAAC DATA
                    elif (self.o.dataSource == "podaac" and 
                            'GET DATA' in url['Type'] and
                            'Download' in url['Description'] and
                            'podaac' in url['URL']):
                        logger.debug(f"PODAAC data URL {url['URL']}")
                        data_url = url['URL']

                    # PO.DAAC md5
                    elif (self.o.dataSource == "podaac" and 
                            'EXTENDED METADATA' in url['Type'] and
                            'Download' in url['Description'] and
                            'md5' in url['Description'] and
                            'podaac' in url['URL']):
                        logger.debug(f"PODAAC md5 URL {url['URL']}")
                        md5_url = url['URL']
                    
                    # Other
                    elif (self.o.dataSource == "other" and 
                            self.o.relatedUrl_type == url['Type'] ):
                        logger.debug(f"Other ({self.o.relatedUrl_type}) URL {url['URL']}")
                        # Skip this URL when other options are defined and don't match
                        if ( (hasattr(self.o, "relatedUrl_descriptionContains") and 
                                all(desc not in url['Description'] for desc in self.o.relatedUrl_descriptionContains))
                              or (hasattr(self.o, "relatedUrl_urlContains") and
                                all(url_c not in url['URL'] for url_c in self.o.relatedUrl_urlContains)) ):
                            logger.debug(f"Skipping {url['URL']} with Description {url['Description']}..." + 
                                " doesn't match relatedUrl_descriptionContains relatedUrl_urlContains options")
                        else:
                            data_url = url['URL']
                
                # Now that above loop is done, should have a data URL and maybe an md5_url
                sumstr = None
                if md5_url and self.o.dataSource == "podaac":
                    try:
                        md5_resp = requests.get(md5_url)
                        md5 = md5_resp.text.split(" ")[0]
                        logger.debug(f"MD5 Checksum: {md5}")
                        sumstr = {"method":"md5", "value":md5}
                    except:
                        logger.warning(f"Could not get MD5 Checksum for {data_url}, posting without it.")

                # finally create the message!
                if data_url:
                    # The message is created using the post_baseUrl and relative path
                    url = urlparse(data_url)
                    self.o.post_baseUrl = url.scheme + "://" + url.netloc + "/"
                    m = sarracenia.Message.fromFileInfo(url.path, self.o)
                    if m:
                        if sumstr:
                            logger.info(f"md5sum is available for {data_url}. Changing identity from {m['identity']} to {sumstr}")
                            m['identity'] = sumstr
                        logger.info(f"new message created for {data_url}")
                        logger.debug(f"new message for {data_url} is\t{m}")
                        gathered_messages.append(m)
                    else:
                        logger.error(f"failed to create message for {data_url}")
                else:
                    logger.debug(f"couldn't find a URL to post")
        
        return gathered_messages



import logging
import os
from sarracenia.flowcb import FlowCB

logger = logging.getLogger(__name__)


class Rxpipe(FlowCB):
    """
      After each file is transferred, write it's name to a named_pipe.
      
      parameter:
              rxpipe_name  -- the path for the named pipe to write the file names to.
    """
    def __init__(self, options):

        super().__init__(options,logger)
        self.o.add_option(option='rxpipe_name', kind='str')

    def on_start(self):
        if not hasattr(self.o, 'rxpipe_name') and self.o.file_rxpipe_name:
            logger.error("Missing rxpipe_name parameter")
            return
        self.rxpipe = open(self.o.rxpipe_name, "w")
        logger.info(f'opened rxpipe_name: {self.o.rxpipe_name} for append')

    def after_work(self, worklist):

        for msg in worklist.ok:
            fname = msg['new_dir'] + os.sep + msg['new_file']
            self.rxpipe.write(fname + '\n')
        self.rxpipe.flush()
        return None



"""
Plugin speedo.py:
    Gives a *speedometer* reading on the messages going through an exchange.
    as this is an after_accept
    Accumulate the number of messages and the bytes they represent over a period of time.

Options:
    msgSpeedoInterval -> how often the speedometer is updated. (default: 5)
    msg_speedo_maxlag  -> if the message flow indicates that messages are 'late', emit warnings. (default 60)

Usage: 
    callback accept.speedo
    msgSpeedoInterval x
    msg_speedo_maxlag y
"""


from sarracenia import timestr2flt, nowflt, naturalSize, naturalTime
import logging
from sarracenia.flowcb import FlowCB
logger = logging.getLogger(__name__)


class Speedo(FlowCB):
    def __init__(self, options):
        """
            set defaults for options.  can be overridden in config file.
        """
        super().__init__(options, logger)

        self.o.add_option('msg_speedo_maxlag', 'count', 60)
        #if hasattr(self.o, 'msg_speedo_maxlag'):
        #    if type(self.o.msg_speedo_maxlag) is list:
        #        self.o.msg_speedo_maxlag = int(self.o.msg_speedo_maxlag[0])
        #else:
        #    self.o.msg_speedo_maxlag = 60
        logger.debug("speedo init: 2 ")

        self.o.add_option('msgSpeedoInterval', 'count', 5)
        #if hasattr(self.o, 'msgSpeedoInterval'):
        #    if type(self.o.msgSpeedoInterval) is list:
        #        self.o.msgSpeedoInterval = int(self.o.msgSpeedoInterval[0])
        #else:
        #    self.o.msgSpeedoInterval = 5

        now = nowflt()
        self.msg_speedo_last = now
        self.msg_speedo_msgcount = 0
        self.msg_speedo_bytecount = 0

    def after_accept(self, worklist):
        for message in worklist.incoming:
            msgtime = timestr2flt(message['pubTime'])
            now = nowflt()
            self.msg_speedo_msgcount = self.msg_speedo_msgcount + 1

            (method, psize, ptot, prem, pno) = message['partstr'].split(',')

            self.msg_speedo_bytecount += int(psize)

            #not time to report yet.
            if self.o.msgSpeedoInterval > now - self.msg_speedo_last:
                continue

            lag = now - msgtime
            msgpersec = self.msg_speedo_msgcount / (now - self.msg_speedo_last)
            bytespersec = self.msg_speedo_bytecount / (now - self.msg_speedo_last)
            logger.info("speedo: %3d messages received: %5.4f msg/s, %4.2f bytes/s, lag: %4.0f s" % (self.msg_speedo_msgcount, msgpersec, bytespersec, lag))

            # If lag is higher than max allowed, emmit a warning
            if lag > self.o.msg_speedo_maxlag:
                logger.warning("speedo: Excessive lag! Messages posted %4.0f s ago" % lag)

            self.msg_speedo_last = now
            self.msg_speedo_msgcount = 0
            self.msg_speedo_bytecount = 0



#!/usr/bin/python3

"""

This downloader corresponds to the poll.mail class.
Fetches emails from a server using the specified protocol,
either pop3 or imap. The imaplib/poplib implementations in
Python use the most secure SSL settings by default: 
PROTOCOL_TLS, OP_NO_SSLv2, and OP_NO_SSLv3.
Compatible with Python 2.7+.

download_email_ingest: a sample do_download option for sr_subscribe.
                        connects to an email server with the provided
                        credentials and posts all new messages.

usage:
        in an sr_subscribe configuration file:

        destination [imap|imaps|pop|pops]://[user[:password]@]host[:port]/
        do_download download_email_ingest

        IMAP over SSL uses 993, POP3 over SSL uses 995
        IMAP unsecured uses 143, POP3 unsecured uses 110

        Full credentials must be in credentials.conf.
        If port is not specified it'll default to the ones above based on protocol/ssl setting.

"""

import poplib, imaplib, datetime, logging, email, urllib

import sarracenia
from sarracenia.flowcb import FlowCB
import sarracenia.identity


logger = logging.getLogger(__name__)

class Mail_ingest(FlowCB):

        def download(self, msg) -> bool:

                #ok, details = self.o.credentials.get(parent.msg.notice.split()[1])
                ok, details = self.o.credentials.get(msg.baseUrl)
                if ok: 
                        setting         = details.url
                        user            = setting.username
                        user = urllib.parse.unquote(user)
                        password        = setting.password
                        password = urllib.parse.unquote(password)
                        server          = setting.hostname
                        protocol        = setting.scheme.lower() 
                        port            = setting.port
                else:
                        logger.error("download_email_ingest: destination has invalid credentials: %s" % msg['baseUrl'])
                        return False

                if not port:
                        if protocol == "imaps":
                                port = 993
                        elif protocol == "pops":
                                port = 995
                        elif protocol == "imap":
                                port = 143
                        else:
                                port = 110

                if "imap" in protocol:
                        if protocol == "imaps":
                                try:
                                        mailman = imaplib.IMAP4_SSL(server, port=port)
                                        mailman.login(user, password)
                                except imaplib.IMAP4.error as e:
                                        logger.error("download_email_ingest imaplib connection error: {}".format(e))
                                        return False

                        elif protocol == "imap":
                                try:
                                        mailman = imaplib.IMAP4(server, port=port)
                                        mailman.login(user, password)
                                except imaplib.IMAP4.error as e:
                                        logger.error("download_email_ingest imaplib connection error: {}".format(e))
                                        return False
                        else: return False

                        mailman.select(mailbox='INBOX')
                        resp, data = mailman.search(None, 'ALL')
                        for index in data[0].split():
                                r, d = mailman.fetch(index, '(RFC822)')
                                msg = d[0][1].decode("utf-8", "ignore") + "\n"
                                logger.info("download_email_ingest downloaded file: %s" % msg['new_dir']+'/'+msg['new_file'])
                                with open(msg['new_dir']+'/'+msg['new_file'], 'w') as f:
                                       f.write(msg)
                                       f.close()
                                if self.o.delete :
                                       mailman.store(index, '+FLAGS', '\\Deleted')

                        mailman.expunge()
                        mailman.close()
                        mailman.logout()
                        return True
                        
                elif "pop" in protocol:
                        if protocol == "pops":
                                try:
                                        mailman = poplib.POP3_SSL(server, port=port)
                                        mailman.user(user)
                                        mailman.pass_(password)
                                except poplib.error_proto as e:
                                        logger.error("download_email_ingest pop3 connection error: {}".format(e))
                                        return False

                        elif protocol == "pop":
                                try:
                                        mailman = poplib.POP3(server, port=port)
                                        mailman.user(user)
                                        mailman.pass_(password)
                                except poplib.error_proto as e:
                                        logger.error("download_email_ingest pop3 connection error: {}".format(e))
                                        return False
                        else: return False
                        # only retrieves msgs that haven't triggered internal pop3 'read' flag
                        numMsgs = len(mailman.list()[1])
                        found = False
                        for index in range(numMsgs):

                                # read the email message
                                email_message=""
                                for line in mailman.retr(index+1)[1]:
                                        email_message += line.decode("utf-8", "ignore") + "\n"
                                msgid = email.message_from_string(email_message).get('Message-ID').strip('<>')

                                # if it worked, write it locally, update message as needed.
                                if msgid == msg['new_file']:                                        
                                    logger.info("download_email_ingest downloaded file: %s" % msg['new_dir']+'/'+msg['new_file'])
                                    
                                    sumalgo = sarracenia.identity.Identity.factory(self.o.identity_method)
                                    sumalgo.set_path(path)
                                    with open(msg['new_dir']+'/'+msg['new_file'], 'w') as f:
                                        sumalgo.update(email_message)
                                        f.write(email_message)
                                        f.close()
                                
                                    message['size'] = len(bytes(email_message,'utf8'))
                                    message['identity'] = { 'method': self.o.identity_method, 'value': sumalgo.value }
                                    if self.o.delete :
                                        mailman.dele(index+1)
                                    found=True
                                    break

                        if not found:
                           logger.error('message not found on server, giving up')

                        mailman.quit()
                        #v2 parent.msg.set_parts()-> sr3? msg.updatePaths( self.o, msg['new_dir'], msg['new_file'] )
                        return True

                else:
                        logger.error("download_email_ingest destination protocol must be one of 'imap/imaps' or 'pop/pops'.")
                        return False



# This file is part of sarracenia.
# The sarracenia suite is Free and is proudly provided by the Government of Canada
# Copyright (C) Her Majesty The Queen in Right of Canada, Environment Canada, 2008-2020
#
# Sarracenia repository: https://github.com/MetPX/sarracenia
# Documentation: https://github.com/MetPX/sarracenia
#
"""
  Place holder... thought this was not needed, then there was a stack trace.

"""
pass



"""
This plugin can be use to add ownership and group of a file in a post message (on_post)
and change the owner/group of products at destination (on_file)

Sample usage:

rootChownMappingFile mapping.conf

flowcb rootchown

Options
-------

If a users/groups differs from the source to the destination, the user can supply a mapping file
which would associate  SRC_UG to DEST_UG.  The filepath is given by giving an absolute path with
option 'mapping_file'. Default value is None, which means give set ownership as for the source user/group.
The 'mapping_file' file format would simply be, a one liner per owner/group

aspymjg:cidx mjg777:ssc_di

here aspymjg:cidx would be the source ownership (source user:group)
and  mjg777:ssc_di  the destination ownership (destination user:group)

"""

import grp
import logging
import os
import pwd
from sarracenia.flowcb import FlowCB

logger = logging.getLogger(__name__)


class Rootchown(FlowCB):
    def __init__(self, options):
        super().__init__(options,logger)
        self.o.declare_option('rootChownMappingFile')
        self.mapping = {}

    def on_start(self):

        if not hasattr(self.o, "rootChownMappingFile"):
            self.o.mapping_file = None
            return True

        mf_path = self.o.mapping_file[0]
        try:
            f = open(mf_path, 'r')
            while True:
                l = f.readline()
                if not l: break
                l2 = l.strip()
                parts = l2.split()
                if len(parts) != 2:
                    logger.error("wrong mapping line %s" % l)
                    continue
                self.mapping[parts[0]] = parts[1]
            f.close()
            logger.info("ROOT_CHOWN mapping_file loaded  %s" % mf_path)
        except:
            logger.error("ROOT_CHOWN problem when parsing %s" % mf_path)

    def after_accept(self, worklist):
        logger.debug("ROOT_CHOWN after_accept")

        for message in worklist.incoming:
            new_dir = message['new_dir']
            new_file = message['new_file']

            # if remove ...

            if 'fileOp' in message and ( ( 'remove' in message['fileOp'] ) or ( 'rename' in message['fileOp'] ) ):
                continue

            # if move ... sr_watch sets new_dir new_file on destination file so we are ok

            # already set ... check for mapping switch

            if 'ownership' in message:
                ug = message['ownership']
                if ug in self.mapping:
                    logger.debug("ROOT_CHOWN mapping from %s to %s" %
                                 (ug, self.mapping[ug]))
                    message['ownership'] = self.mapping[ug]

            # need to add ownership in message

            try:
                local_file = new_dir + os.sep + new_file

                s = os.lstat(local_file)
                username = pwd.getpwuid(s.st_uid).pw_name
                group = grp.getgrgid(s.st_gid).gr_name

                ug = "%s:%s" % (username, group)

                # check for mapping switch
                if ug in self.mapping:
                    logger.debug("ROOT_CHOWN mapping from %s to %s" %
                                 (ug, self.mapping[ug]))
                    ug = self.mapping[ug]

                message['ownership'] = ug
                logger.debug("ROOT_CHOWN set ownership field %s" %
                             message['ownership'])

            except:
                logger.error("ROOT_CHOWN could not set ownership  %s" %
                             local_file)
                #FIXME should we do worklist.reject here?

    def after_work(self, worklist):
        logger.debug("ROOT_CHOWN after_work")

        for message in worklist.ok:
            # the message does not have the requiered info

            if not 'ownership' in message:
                logger.info("ROOT_CHOWN no ownership field in msg_headers")
                # FIXME worklist.reject here?
                continue

            # it does, check for mapping

            ug = message['ownership']
            if ug in self.mapping:
                logger.debug("received ownership %s mapped to %s" %
                             (ug, self.mapping[ug]))
                ug = self.mapping[ug]

            # try getting/setting ownership info to local_file

            local_file = message['new_dir'] + os.sep + message['new_file']

            try:
                parts = ug.split(':')
                username = parts[0]
                group = parts[1]

                uid = pwd.getpwnam(username).pw_uid
                gid = grp.getgrnam(group).pw_gid

                os.chown(local_file, uid, gid)
                logger.info("ROOT_CHOWN set ownership %s to %s" %
                            (ug, local_file))
                #FIXME not sure if we add to worklist.ok here

            except:
                logger.error("ROOT_CHOWN could not set %s to %s" %
                             (ug, local_file))



import logging
from sarracenia.flowcb import FlowCB

logger = logging.getLogger(__name__)


class Data(FlowCB):
    """
       duplicate suppression based on data alone. Overrides the path used for lookups
       in the cache so that all files have the same name, and so if the checksum
       is the same, regardless of file name, it is considered a duplicate.
    """
    def after_accept(self, worklist):
        for m in worklist.incoming:
            if not 'nodupe_override' in m:
                m['_deleteOnPost'] |= set(['nodupe_override'])
                m['nodupe_override'] = {}
            m['nodupe_override']['path'] = 'data'



# This file is part of sarracenia.
# The sarracenia suite is Free and is proudly provided by the Government of Canada
# Copyright (C) Her Majesty The Queen in Right of Canada, 2008-2021
#
# Sarracenia repository: https://github.com/MetPX/sarracenia
# Documentation: https://github.com/MetPX/sarracenia
#
########################################################################
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; version 2 of the License.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
#

import ftplib, os, subprocess, sys, time
import logging
from sarracenia.transfer import Transfer
from sarracenia.transfer import alarm_cancel, alarm_set, alarm_raise
from urllib.parse import unquote

logger = logging.getLogger(__name__)


class Ftp(Transfer):
    """
    File Transfer Protocol (FTP)  ( https://datatracker.ietf.org/doc/html/rfc959 )
    sarracenia transfer protocol subclass supports/uses additional custom options:

    * accelFtpputCommand (default: '/usr/bin/ncftpput %s %d' )
    * accelFtpgetCommand (default: '/usr/bin/ncftpget %s %d' )

    built using: ftplib ( https://docs.python.org/3/library/ftplib.html )
    """
    def __init__(self, proto, options):

        super().__init__(proto, options)

        self.o.add_option('accelFtpputCommand', 'str',
                          '/usr/bin/ncftpput %s %d')
        self.o.add_option('accelFtpgetCommand', 'str',
                          '/usr/bin/ncftpget %s %d')

        logger.debug("sr_ftp __init__")
        self.connected = False
        self.ftp = None
        self.details = None
        self.batch = 0

    def registered_as():
        return ['ftp']

    # cd
    def cd(self, path):
        logger.debug("sr_ftp cd %s" % path)

        alarm_set(self.o.timeout)
        self.ftp.cwd(self.originalDir)
        self.ftp.cwd(path)
        self.pwd = path
        alarm_cancel()

    def cd_forced(self, perm, path):
        logger.debug("sr_ftp cd_forced %d %s" % (perm, path))

        # try to go directly to path

        alarm_set(self.o.timeout)
        self.ftp.cwd(self.originalDir)
        try:
            self.ftp.cwd(path)
            alarm_cancel()
            return
        except:
            pass
        alarm_cancel()

        # need to create subdir

        subdirs = path.split("/")
        if path[0:1] == "/": subdirs[0] = "/" + subdirs[0]

        for d in subdirs:
            if d == '': continue
            # try to go directly to subdir
            try:
                alarm_set(self.o.timeout)
                self.ftp.cwd(d)
                alarm_cancel()
                continue
            except:
                pass

            # create
            alarm_set(self.o.timeout)
            self.ftp.mkd(d)
            alarm_cancel()

            # chmod
            alarm_set(self.o.timeout)
            self.ftp.voidcmd('SITE CHMOD ' + "{0:o}".format(perm) + ' ' + d)
            alarm_cancel()

            # cd
            alarm_set(self.o.timeout)
            self.ftp.cwd(d)
            alarm_cancel()

    # check_is_connected

    def check_is_connected(self):
        logger.debug("sr_ftp check_is_connected")

        if self.ftp == None: return False
        if not self.connected: return False

        if self.sendTo != self.o.sendTo:
            self.close()
            return False

        self.batch = self.batch + 1
        if self.batch > self.o.batch:
            self.close()
            return False

        # really connected
        try:
            cwd = self.getcwd()
        except:
            self.close()
            return False

        return True

    # chmod
    def chmod(self, perm, path):
        logger.debug("sr_ftp chmod %s %s" % (str(perm), path))
        alarm_set(self.o.timeout)
        self.ftp.voidcmd('SITE CHMOD ' + "{0:o}".format(perm) + ' ' + path)
        alarm_cancel()

    # close
    def close(self):
        logger.debug("sr_ftp close")

        old_ftp = self.ftp

        self.init()

        try:
            alarm_set(self.o.timeout)
            old_ftp.quit()
        except:
            pass
        alarm_cancel()

    # connect...
    def connect(self):
        logger.debug("sr_ftp connect %s" % self.o.sendTo)

        self.connected = False
        self.sendTo = self.o.sendTo

        if not self.credentials(): return False

        # timeout alarm 100 secs to connect
        alarm_set(self.o.timeout)
        try:
            expire = -999
            if self.o.timeout: expire = self.o.timeout
            if self.port == '' or self.port == None: self.port = 21

            if not self.tls:
                ftp = ftplib.FTP()
                ftp.encoding = 'utf-8'
                ftp.connect(self.host, self.port, timeout=expire)
                ftp.login(self.user, unquote(self.password))
            else:
                # ftplib supports FTPS with TLS
                ftp = ftplib.FTP_TLS(self.host,
                                     self.user,
                                     unquote(self.password),
                                     timeout=expire)
                ftp.encoding = 'utf-8'
                if self.prot_p: ftp.prot_p()
                # needed only if prot_p then set back to prot_c
                #else          : ftp.prot_c()

            ftp.set_pasv(self.passive)

            self.originalDir = '.'

            try:
                self.originalDir = ftp.pwd()
            except:
                logger.warning("Unable to ftp.pwd")
                logger.debug('Exception details: ', exc_info=True)

            self.pwd = self.originalDir

            self.connected = True

            self.ftp = ftp

            #alarm_cancel()
            return True

        except:
            logger.error("Unable to connect to %s (user:%s)" %
                         (self.host, self.user))
            logger.debug('Exception details: ', exc_info=True)

        alarm_cancel()
        return False

    # credentials...
    def credentials(self):
        logger.debug("sr_ftp credentials %s" % self.sendTo)

        try:
            ok, details = self.o.credentials.get(self.sendTo)
            if details: url = details.url

            self.host = url.hostname
            self.port = url.port
            self.user = url.username
            self.password = url.password

            self.passive = details.passive
            self.binary = details.binary
            self.tls = details.tls
            self.prot_p = details.prot_p

            return True

        except:
            logger.error(
                "sr_ftp/credentials: unable to get credentials for %s" %
                self.sendTo)
            logger.debug('Exception details: ', exc_info=True)

        return False

    # delete
    def delete(self, path):
        logger.debug("sr_ftp rm %s" % path)
        alarm_set(self.o.timeout)
        # if delete does not work (file not found) run pwd to see if connection is ok
        try:
            self.ftp.delete(path)
        except:
            d = self.ftp.pwd()
        alarm_cancel()

    # get
    def get(self,
            msg,
            remote_file,
            local_file,
            remote_offset=0,
            local_offset=0,
            length=0):
        logger.debug("sr_ftp get %s %s %d" %
                     (remote_file, local_file, local_offset))

        # open local file
        dst = self.local_write_open(local_file, local_offset)

        # initialize sumalgo
        if self.sumalgo: self.sumalgo.set_path(remote_file)

        # download
        self.write_chunk_init(dst)
        if self.binary:
            self.ftp.retrbinary('RETR ' + remote_file, self.write_chunk,
                                self.o.bufsize)
        else:
            self.ftp.retrlines('RETR ' + remote_file, self.write_chunk)
        rw_length = self.write_chunk_end()

        # close
        self.local_write_close(dst)

        return rw_length

    def getAccelerated(self, msg, remote_file, local_file, length=0):

        base_url = msg['baseUrl']
        if base_url[-1] == '/':
            base_url = base_url[0:-1]
        arg1 = base_url + self.pwd + os.sep + remote_file
        arg1 = arg1.replace(' ', '\ ')
        arg2 = local_file

        cmd = self.o.accelFtpgetCommand.replace('%s', arg1)
        cmd = cmd.replace('%d', arg2).split()

        logger.info("accel_ftp:  %s" % ' '.join(cmd))
        p = subprocess.Popen(cmd)
        p.wait()
        if p.returncode != 0:
            return -1
        sz = os.stat(arg2).st_size
        return sz

    # getcwd
    def getcwd(self):
        alarm_set(self.o.timeout)
        pwd = self.ftp.pwd()
        alarm_cancel()
        return pwd

    # ls
    def ls(self):
        logger.debug("sr_ftp ls")
        self.entries = {}
        alarm_set(self.o.timeout)
        self.ftp.retrlines('LIST', self.line_callback)
        alarm_cancel()
        logger.debug("sr_ftp ls = (size: %d) %s ..." % (len(self.entries), str(self.entries)[0:255]))
        return self.entries

    # line_callback: entries[filename] = 'stripped_file_description'
    def line_callback(self, iline):
        #logger.debug("sr_ftp line_callback %s" % iline)

        alarm_cancel()

        oline = iline
        oline = oline.strip('\n')
        oline = oline.strip()
        oline = oline.replace('\t', ' ')
        opart1 = oline.split(' ')
        opart2 = []

        for p in opart1:
            if p == '': continue
            opart2.append(p)

        # else case is in the event of unlikely race condition

        # on linux, there are 8 fields, with spaces, perhaps more...
        if len(opart2) > 7:
            # university of Wisconsin as an ftp server that has an extra auth field.
            if opart2[4].isnumeric(): # normal linux case.
                fil = ' '.join(opart2[8:])
            else:  # U. Wisconsin case.
                fil = ' '.join(opart2[9:])
        else:
            # guess it is on windows... 
            fil = ' '.join(opart2[3:])

        line = ' '.join(opart2)

        self.entries[fil] = line

        alarm_set(self.o.timeout)

    # mkdir
    def mkdir(self, remote_dir):
        logger.debug("sr_ftp mkdir %s" % remote_dir)
        alarm_set(self.o.timeout)
        self.ftp.mkd(remote_dir)
        alarm_cancel()
        alarm_set(self.o.timeout)
        self.ftp.voidcmd('SITE CHMOD ' +
                         "{0:o}".format(self.o.permDirDefault) + ' ' +
                         remote_dir)
        alarm_cancel()

    # put
    def put(self,
            msg,
            local_file,
            remote_file,
            local_offset=0,
            remote_offset=0,
            length=0):
        logger.debug("sr_ftp put %s %s" % (local_file, remote_file))

        # open
        src = self.local_read_open(local_file, local_offset)

        # upload
        self.write_chunk_init(None)
        if self.binary:
            self.ftp.storbinary("STOR " + remote_file, src, self.o.bufsize,
                                self.write_chunk)
        else:
            self.ftp.storlines("STOR " + remote_file, src, self.write_chunk)
        rw_length = self.write_chunk_end()

        # close
        self.local_read_close(src)

        return rw_length

    def putAccelerated(self, msg, local_file, remote_file, length=0):

        dest_baseUrl = self.o.sendTo
        if dest_baseUrl[-1] == '/':
            dest_baseUrl = dest_baseUrl[0:-1]
        arg2 = dest_baseUrl + msg['new_dir'] + os.sep + remote_file
        arg2 = arg2.replace(' ', '\ ')
        arg1 = local_file

        cmd = self.o.accelFtpputCommand.replace('%s', arg1)
        cmd = cmd.replace('%d', arg2).split()

        logger.info("accel_ftp:  %s" % ' '.join(cmd))
        p = subprocess.Popen(cmd)
        p.wait()
        if p.returncode != 0:
            return -1
        # FIXME: faking success... not sure how to check really.
        sz = int(msg['size'])
        return sz

    # rename
    def rename(self, remote_old, remote_new):
        logger.debug("sr_ftp rename %s %s" % (remote_old, remote_new))
        alarm_set(self.o.timeout)
        self.ftp.rename(remote_old, remote_new)
        alarm_cancel()

    # rmdir
    def rmdir(self, path):
        logger.debug("sr_ftp rmdir %s" % path)
        alarm_set(self.o.timeout)
        self.ftp.rmd(path)
        alarm_cancel()

    # umask
    def umask(self):
        logger.debug("sr_ftp umask")
        alarm_set(self.o.timeout)
        self.ftp.voidcmd('SITE UMASK 777')
        alarm_cancel()



"""
wmo2msc.py is an on_message plugin script to convert WMO bulletins on local disk
to MSC internal format in an alternate tree.  It is analogous to Sundew's 'bulletin-file'.
Meant to be called as an sr_shovel plugin.

It prints an output line:

wmo2msc: <input_file> -> <output_file> (<detected format>)

usage:

Use the directory setting to know the root of tree where files are placed.
FIXME: well, likely what you really what is something like::

     <date>/<source>/dir1/dir2/dir3

     <date>/<source>/dir1/dir2/newdir4/...

     -- so Directory doesn't cut it.

In a sr_shovel configuration:: 

    directory /.... 
    callback filter.wmo2msc


Parameters:

* filter_wmo2msc_replace_dir  old,new

* filter_wmo2msc_uniquify hash|time|anything else
  - whether to add a string in addition to the AHL to make the filename unique.
  - hash - means apply a hash, so that the additional string is content based.
  - if time, add a suffix _YYYYMMDDHHMMSS_99999 which ensures file name uniqueness.
  - otherwise, no suffix will be added.
  - default: hash

* filter_wmo2msc_convert on|off
  if on, then traditional conversion to MSC-BULLETINS is done as per TANDEM/APPS & MetPX Sundew
  this involves \n as termination character, and other charater substitutions.

* filter_wmo2msc_tree  on|off
  if tree is off, files are just placed in destination directory.
  if tree is on, then the file is placed in a subdirectory tree, based on
  the WMO 386 AHL::

         TTAAii CCCC YYGGgg  ( example: SACN37 CWAO 300104 )
 
         TT = SA - surface observation.
         AA = CN - Canada ( but the AA depends on TT value, in many cases not a national code. )
         ii = 37 - a number.. there are various conventions, they are picked to avoid duplication.
     
  The first line of the file is expected to contain an AHL. and when we build a tree
  from it, we build it as follows::

     TT/CCCC/GG/TTAAii_CCCC_YYGGgg_<uniquify>

  assuming tree=on, uniquify=hash:

     SA/CWAO/01/SACN37_CWAO_300104_1c699da91817cc4a84ab19ee4abe4e22


NOTE: Look at the end of the file for SUPPLEMENTARY INFORMATION 
      including hints about debugging.

"""

import sys
import os
import re
import time
import hashlib
import logging
import random
from sarracenia.flowcb import FlowCB

logger = logging.getLogger('__name__')


class Wmo2msc(FlowCB):
    def __init__(self, options):
        super().__init__(options,logger)

        self.o.add_option( 'filter_wmo2msc_uniquify', 'str', 'hash' )
        self.o.add_option( 'filter_wmo2msc_replace_dir', 'str' )
        self.o.add_option( 'filter_wmo2msc_treeify', 'flag', True )
        self.o.add_option( 'filter_wmo2msc_convert', 'flag', True )

        if not hasattr(self.o, 'filter_wmo2msc_replace_dir'):
            logger.error("filter_wmo2msc_replace_dir setting is mandatory")
            return

        (self.o.filter_olddir, self.o.filter_newdir) = self.o.filter_wmo2msc_replace_dir.split(',')

        logger.info("filter_wmo2msc old-dir=%s, newdir=%s" %
                    (self.o.filter_olddir, self.o.filter_newdir))

        self.trimre = re.compile(b" +\n")

    def replaceChar(self, oldchar, newchar):
        """
           replace all occurrences of oldchar by newchar in the the message byte stream.
           started as a direct copy from sundew of routine with same name in bulletin.py
           - the storage model is a bit different, we store the entire message as one bytearray.
           - sundew stored it as a series of lines, so replaceChar implementation changed.

        """
        self.bintxt = self.bintxt.replace(bytearray(oldchar, 'latin_1'),
                                          bytearray(newchar, 'latin_1'))

    def doSpecificProcessing(self):
        """doSpecificProcessing()

           Modify bulletins received from Washington via the WMO socket protocol.
           started as a direct copy from sundew of routine with same name in bulletinManagerWmo.py
      
           - encode/decode, and binary stuff came because of python3
        """

        ahl2 = self.bulletin[0][:2].decode('ascii')
        ahl4 = self.bulletin[0][:4].decode('ascii')

        if ahl2 in [
                'SD', 'SO', 'WS', 'SR', 'SX', 'FO', 'WA', 'AC', 'FA', 'FB',
                'FD'
        ]:
            self.replaceChar('\x1e', '')

        if ahl2 in ['SR', 'SX']:
            self.replaceChar('~', '\n')

        if ahl2 in ['UK']:
            self.replaceChar('\x01', '')

        if ahl4 in ['SICO']:
            self.replaceChar('\x01', '')

        if ahl2 in ['SO', 'SR']:
            self.replaceChar('\x02', '')

        if ahl2 in ['SX', 'SR', 'SO']:
            self.replaceChar('\x00', '')

        if ahl2 in ['SX']:
            self.replaceChar('\x11', '')
            self.replaceChar('\x14', '')
            self.replaceChar('\x19', '')
            self.replaceChar('\x1f', '')

        if ahl2 in ['SR']:
            self.replaceChar('\b', '')
            self.replaceChar('\t', '')
            self.replaceChar('\x1a', '')
            self.replaceChar('\x1b', '')
            self.replaceChar('\x12', '')

        if ahl2 in ['FX']:
            self.replaceChar('\x10', '')
            self.replaceChar('\xf1', '')

        if ahl2 in ['WW']:
            self.replaceChar('\xba', '')

        if ahl2 in ['US']:
            self.replaceChar('\x18', '')

        if ahl4 in ['SXUS', 'SXCN', 'SRCN']:
            self.replaceChar('\x7f', '?')

        if ahl4 in ['SRCN']:
            self.replaceChar('\x0e', '')
            self.replaceChar('\x11', '')
            self.replaceChar('\x16', '')
            self.replaceChar('\x19', '')
            self.replaceChar('\x1d', '')
            self.replaceChar('\x1f', '')

        if ahl4 in ['SXVX', 'SRUS', 'SRMT']:
            self.replaceChar('\x7f', '')

        self.replaceChar('\r', '')

        if ahl2 in ['SA', 'SM', 'SI', 'SO', 'UJ', 'US', 'FT']:
            self.replaceChar('\x03', '')

        #trimming of trailing blanks.
        lenb = len(self.bintxt)
        self.bintxt = self.trimre.sub(b"\n", self.bintxt)
        if len(self.bintxt) < lenb:
            print('Trimmed %d trailing blanks!' % (lenb - len(self.bintxt)))

    def after_accept(self, worklist):
        new_incoming = []
        for message in worklist.incoming:
            if message['baseUrl'] != 'file:':
                logger.error(
                    'filter_wmo2msc needs local files invalid url: %s ' %
                    (message['baseUrl'] + message['relPath']))
                worklist.rejected.append(message)
                continue

            input_file = message['relPath']

            # read once to get headers and type.

            logger.debug('filter_wmo2msc reading file: %s' % (input_file))

            with open(input_file, 'rb') as s:
                self.bulletin = [s.readline(), s.read(4)]

            AHLfn = (self.bulletin[0].replace(b' ',
                                              b'_').strip()).decode('ascii')

            if len(AHLfn) < 18:
                logger.error(
                    'filter_wmo2msc: not a WMO bulletin, malformed header: (%s)'
                    % (AHLfn))
                worklist.rejected.append(message)
                continue

            # read second time for the body in one string.
            with open(input_file, 'rb') as s:
                self.bintxt = s.read()

            logger.debug('filter_wmo2msc read twice: %s ' % (input_file))

            # Determine file format (fmt) and apply transformation.
            if self.bulletin[1].lstrip()[:4] in ['BUFR', 'GRIB', '\211PNG']:
                fmt = 'wmo-binary'
                self.replaceChar('\r', '')
            elif self.bulletin[0][:11] in ['SFUK45 EGRR']:
                # This file is encoded in an indecipherably non-standard format.
                fmt = 'unknown-binary'

                #self.replaceChar('\r','',2) replace only the first 2 carriage returns.
                self.bintxt = self.bintxt.replace(bytearray('\r', 'latin_1'),
                                                  bytearray('', 'latin_1'), 2)
            else:
                fmt = 'wmo-alphanumeric'
                if self.o.filter_wmo2msc_convert:
                    self.doSpecificProcessing()

            # apply 'd' checksum (md5)

            s = hashlib.md5()
            s.update(self.bintxt)
            sumstr = ''.join(format(x, '02x') for x in s.digest())

            # Determine local file name.
            if self.o.filter_wmo2msc_uniquify in ['time']:

                AHLfn += '_' + time.strftime( "%Y%m%d%H%M%S", time.gmtime(time.time()) ) + \
                         '_%05d' % random.randint(0,9999)
            elif self.o.filter_wmo2msc_uniquify in ['hash']:
                #AHLfn += '_%s' % ''.join( format(x, '02x') for x in s.digest() )
                AHLfn += '_' + sumstr

            if self.o.filter_wmo2msc_treeify:
                d = os.path.dirname(input_file)
                logger.debug('filter_wmo2msc check %s start match: %s' %
                             (d, self.o.filter_olddir))
                d = d.replace(self.o.filter_olddir, self.o.filter_newdir)
                logger.debug('filter_wmo2msc check %s after replace' % (d))
                if not os.path.isdir(d):
                    os.makedirs(d, self.o.permDirDefault, True)

                d = d + os.sep + self.bulletin[0][0:2].decode('ascii')
                d = d + os.sep + self.bulletin[0][7:11].decode('ascii')
                logger.debug('filter_wmo2msc check %s' % (d))
                if not os.path.isdir(d):
                    os.makedirs(d, self.o.permDirDefault, True)

                d = d + os.sep + self.bulletin[0][14:16].decode('ascii')
                logger.debug('filter_wmo2msc check %s' % (d))
                if not os.path.isdir(d):
                    os.makedirs(d, self.o.permDirDefault, True)

                local_file = d + os.sep + AHLfn
            else:
                local_file = self.o.currentDir + os.sep + AHLfn

            # write the data.
            fileOK = False

            if not self.o.filter_wmo2msc_convert:
                try:
                    os.link(input_file, local_file)
                    fileOK = True
                except:
                    pass

            if self.o.filter_wmo2msc_convert or not fileOK:
                d = open(local_file, 'wb+')
                d.write(self.bintxt)
                d.close()

            logger.debug('filter_wmo2msc %s -> %s (%s)' %
                         (input_file, local_file, fmt))

            # set how the file will be announced

            baseDir = self.o.base_dir
            if baseDir == None: baseDir = self.o.post_base_dir

            relPath = local_file
            if baseDir != None: relPath = local_file.replace(baseDir, '')

            baseUrl = 'file:'
            # from tolocal.py if used
            if 'savedUrl' in message.keys(): baseUrl = message['savedUrl']
            # from tolocalfile.py if used
            if 'saved_baseUrl' in message.keys():
                baseUrl = message['saved_baseUrl']

            relPath = relPath.replace('//', '/')
            logger.debug('filter_wmo2msc relPath %s' % relPath)

            message['set_topic'](self.o.topic_prefix, relPath)
            #message['set_notice'](baseUrl, relPath)
            message['baseUrl'] = baseUrl
            message['relPath'] = relPath
            new_incoming.append(message)
        worklist.incoming = new_incoming


"""

SUPPLEMENTARY INFORMATION

SUNDEW COMPATIBILITY:   This script is the sarra version of a 'bulletin-file' receiver.

dod_filter_wmo2msc.py is a do_download plugin script used with sr_sarra to convert
World Meteorological Organisation (WMO) standard (See WMO-386 and WMO-306) bulletins 
into the similar but different internal format used by the Meteorological Service of 
Canada (MSC.)

transformation of bulletins is based on detecting the format by reading the first 
line of the file, and the first four bytes of after the first line.  detected format 
is one of:
   wmo-binary: GRIB, BUFR, or PNG, which require carriage returns to be removed ?
   wmo-alpha:  extensive filtering.
   unknown-binary: unknown format, remove only carriage returns from AHL.

The file is read entirely into memory as the WMO standard specifies a maximum message
size of 500,000 bytes with no segmentation and re-assembly being ruled out.

The output files are named based based on the Abbreviated Header Line (AHL)
from first line of each input file. 

It operates on local files.  One subscribes to a source messages that are already 
downloaded, then this 'download' filter produces a second tree of converted bulletins.

STANDALONE DEBUGGING:
Instead of being used purely as a plugin, the script can also be invoked directly.
In that case, if given no arguments, it will read the current working directory,
looking for files that start with a digit (which is what NWS feed provides.)
and feed those file for processing.

To process a particular files, go into the directory containing the file, and supply
the file names as an arguments.


MSC Format description:

Given a file containing a single WMO 386 / WMO 306 conformant encoded bulletin,
produce an MSC formatted one. Most obvious change is the line termination, but there
are other subtle differences learned through pain, over a year or two of parallel testing.

The MSC format is simply the output of the (late lamented Tandem computer system
which was the bulletin switch from the 80's until 2007) The internal 'AM' circuits
used by the Meteorological Service of Canada used only line feed for termination, as
is standard in Unix/Linux land, and not two carriage returns followed by a line feed required
by the ancient tomes of the WMO.

This processing was determined by blackbox reverse engineering for MetPX sundew in the mid-2000's.
when a study was done over a few days in the mid 2000's, it was determined that approximately 
6% of traffic on the GTS is carriage returns, which seems sad, but the formats are too entrenched 
to be changed.  

Hopefully identical logic has been ported to Sarracenia for use as a sarra plugin in 2017.

Adaptation of sundew code to sarracenia by Peter Silva - 2017/01 

"""



"""
Retrieves file keys in the NEXRAD US Weather Radar public dataset.
Can be adjusted slightly for other public datasets on the AWS S3 platform.
Compatible with Python 3.5+.

sarracenia/flowcb/poll/nexrad.py: a sample gather option for sr3 poll.
	        connects to a public aws s3 bucket and posts all the keys
	        corresponding to files uploaded since it last woke up 

usage:
	in an sr_poll configuration file:
	poll_nexrad_day [YYYY-MM-DD]
	callback poll.nexrad
	
	If poll_nexrad_day is set, it'll post all the keys from that day. Specify a day between 1991-06-05 
	to now (if the current date is specified, it'll post all the keys from today that have been posted
	so far). If not set, it'll post any files uploaded in the last minute (sleep must be set to 51 in 
	the config) and continuously check every minute if a file has been uploaded in the last minute. It 
	takes about ~5 minutes from the timestamp on the file to when it actually gets updated in the bucket,
	and ~9s for the API to make all the requests to check for keys, so the counter starts at now - 7 
	minutes and sleep must be ~51s to work properly.

status: working.
"""

import boto3
from botocore import UNSIGNED
from botocore.client import Config
import datetime
import logging
import paramiko
import sarracenia
from sarracenia.flowcb import FlowCB
import urllib.request

logger = logging.getLogger(__name__)


class Nexrad(FlowCB):
    def __init__(self, options):

        super().__init__(options,logger)
        self.o.add_option('poll_nexrad_day', 'str', "")
        self.minutetracker = datetime.datetime.utcnow() + datetime.timedelta(
            minutes=-7)

    def poll(self):

        # If a valid day wasn't specified, it just won't return any keys or throw an error here ¯\_(ツ)_/¯
        day = self.o.poll_nexrad_day
        if day:
            daypts = day.split('-')
            YYYY = daypts[0].zfill(2)
            MM = daypts[1].zfill(2)
            DD = daypts[2].zfill(2)

        # Gets the list of active weather station ICAOs from
        # https://www.aviationweather.gov/docs/metar/stations.txt
        # According to the website's disclaimer:
        #		" This server is available 24 hours a day, seven days a week.
        #		  Timely delivery of data and products from this server through
        # the Internet is not guaranteed."
        # Considered the primary resource for up-to-date ICAO information.
        # Currently only scrapes US weather station ICAOs, but can be adjusted to pull
        # from Canada/worldwide sites.
        ICAOs = set()
        with urllib.request.urlopen(
                'https://www.aviationweather.gov/docs/metar/stations.txt'
        ) as f:
            lines = f.readlines()
            self.metrics['transferRxBytes'] += len(lines)
            for line in lines:
                line = line.decode("utf-8", "ignore")
                if len(line) > 80:
                    if line.endswith("US\n") and line[65] == 'X':
                        if line[20:24] != "    ": ICAOs.add(line[20:24])

        # Not all sites from the Nexrad data set are covered from the official source
        #(some foreign US bases are included in the NEXRAD dataset), so add the missing ones from this list:
        # https://www.roc.noaa.gov/WSR88D/Program/NetworkSites.aspx
        ICAOs.update(['RKJK', 'PAEC', 'RODN', 'RKSG', 'KGRK'])

        # And for some reason FOP1/NOP4 show up in the dataset, doesn't correlate to active ICAOs though
        ICAOs.update(['FOP1', 'NOP4'])
        # As of 2018/07 this set has 161 elements

        keys = []
        keysizes = []
        s3 = boto3.client('s3',
                          region_name='us-east-1',
                          config=Config(signature_version=UNSIGNED,
                                        retries={'max_attempts': 3}))
        if day:
            # Takes all files uploaded on the specified day. Takes ~161 API calls = ~9s. A full day's worth
            # of keys is around ~37000. Meant to be run once to grab a day's worth of keys. Works as long
            # as a day's worth on a single ICAO doesn't exceed 1000 (the API limit, and as of 2018/07 each
            # ICAO uploads about ~300).
            for station in ICAOs:
                try:
                    for obj in s3.list_objects(Bucket='noaa-nexrad-level2',
                                               Prefix=YYYY + '/' + MM + '/' +
                                               DD + '/' + station +
                                               '/')['Contents']:
                        keys.append(obj['Key'])
                        keysizes.append(obj['Size'])
                except KeyError as e:
                    continue

        else:
            # Forces sleep value to be a minute. Takes ~161 API calls, runs in ~9s.
            self.minutetracker = self.minutetracker + datetime.timedelta(
                minutes=1)
            YYYY = str(self.minutetracker.year)
            MM = str(self.minutetracker.month).zfill(2)
            DD = str(self.minutetracker.day).zfill(2)
            HH = str(self.minutetracker.hour).zfill(2)
            mm = str(self.minutetracker.minute).zfill(2)
            logger.info("Date: %s" % self.minutetracker)

            for station in ICAOs:
                try:
                    for obj in s3.list_objects(
                            Bucket='noaa-nexrad-level2',
                            Prefix=YYYY + '/' + MM + '/' + DD + '/' + station +
                            '/' + station + YYYY + MM + DD + '_' + HH +
                            mm)['Contents']:
                        keys.append(obj['Key'])
                        keysizes.append(obj['Size'])
                except KeyError as e:
                    continue

        # To download a file, you'd append the key to the end of the bucket link, e.g:
        # https://s3.amazonaws.com/noaa-nexrad-level2/2018/06/09/KARX/KARX20180609_000228_V06
        i = 0
        gathered_messages = []
        while i < len(keys):
            logger.info("poll_nexrad: key received {}".format(keys[i]))

            fakeStat = paramiko.SFTPAttributes()
            fakeStat.st_size = keysizes[i]

            m = sarracenia.Message.fromFileInfo(keys[i], self.o, fakeStat)
            gathered_messages.append(m)
            i += 1
        logger.info( f"Done, {len(gathered_messages)} messages to post" )
        return gathered_messages



"""
 A sample on_part plugin to perform virus scanning, using the ClamAV engine.

 requires a clamd binding package to be installed. On debian derived systems:: 

    sudo apt-get install python3-pyclamd

 on others::

    pip3 install pyclamd

 author: Peter Silva
"""

import logging
import os
import stat
import time

from sarracenia import nowflt
import sarracenia
from sarracenia.flowcb import FlowCB

#
# Support for features inventory mechanism.
#
from sarracenia.featuredetection import features

features['clamd'] = { 'modules_needed': [ 'pyclamd' ], 'Needed': True,
        'lament' : 'cannot use clamd to av scan files transferred',
        'rejoice' : 'can use clamd to av scan files transferred' }

try:
    import pyclamd
    features['clamd']['present'] = True
except:
    features['clamd']['present'] = False



logger = logging.getLogger(__name__)


class Clamav(FlowCB):
    """
       Invoke ClamAV anti-virus scanning on files as they pass through a data pump.
       when it is invoked depends on the component it is used from.

       from a sender, post, or poll, the scan should stop processing prior
       to the transfer.

       for other components, subsscribers that download, it needs to take place
       after downloading.

    """
    def __init__(self, options) -> None:

        super().__init__(options,logger)

        self.metric_scanned = 0
        self.metric_hits = 0
       
        if sarracenia.features['pyclamd']['present']:
            import pyclamd
            self.av = pyclamd.ClamdAgnostic()
            print("clam_scan on_part plugin initialized")

    def avscan_hit(self, scanfn) -> bool:

        # worried about how long the scan will take.
        start = nowflt()
        virus_found = self.av.scan_file(scanfn)
        end = nowflt()
        self.metric_scanned += 1

        if virus_found:
            logger.error(
                "part_clamav_scan took %g not forwarding, virus detected in %s"
                % (end - start, scanfn))
            self.metric_hits += 1
            return False

        logger.info("part_clamav_scan took %g seconds, no viruses in %s" %
                    (end - start, scanfn))
        return True

    def after_accept(self, worklist) -> None:
        if self.o.component in ['sender', 'post', 'watch']:
            new_incoming = []
            for m in worklist.incoming:
                scanfn = m['new_dir'] + os.sep + m['new_file']
                logger.info(f'scanning: {scanfn}')
                if self.avscan_hit(scanfn):
                    worklist.rejected.append(m)
                else:
                    new_incoming.append(m)
            worklist.incoming = new_incoming

    def after_work(self, worklist) -> None:
        if self.o.component in ['subscribe', 'sarra']:
            new_ok = []
            for m in worklist.ok:
                scanfn = m['new_dir'] + os.sep + m['new_file']
                logger.info(f'scanning: {scanfn}')
                if self.avscan_hit(scanfn):
                    worklist.rejected.append(m)
                else:
                    new_ok.append(m)
            worklist.ok = new_ok

    def on_housekeeping(self):
        logger.info(
            f'files scanned {self.metric_scanned}, hits: {self.metric_hits} ')
        self.metric_scanned = 0
        self.metric_hits = 0



import datetime
import os
import os.path
import re
import sarracenia.config
import sys

keep_count=5

print( 'EXPERIMENTAL log rotation tool to deal with a bug on windows' )
print( 'bug details: https://github.com/MetPX/sarracenia/issues/785' )
print( 'usage: accept an integer argument: number of old logs to keep, defaults to $keep_count' )
print( 'it does a log rotation each time it is called, if there is a current log to rotate')

if len(sys.argv) > 1:
    print(f" logs to keep: {sys.argv[1]}")
    keep_count=int(sys.argv[1])


timesuffix=str(datetime.datetime.now()).replace(' ','_').replace(':','_')[0:19]
logdir=sarracenia.config.user_cache_dir('sr3','MetPX') + os.sep + 'log'
print( f"log directory is: {logdir}")
current_log_pattern=re.compile(".*\.log$")
current_logs=[]
old_to_delete=[]
old_logs=[]
current_log_root=None
if not os.path.exists(logdir) or not os.path.isdir(logdir):
    print( "log dir does not exist")
    sys.exit(1)
existing_files=sorted(os.listdir(logdir))
os.chdir(logdir)
for l in existing_files:
    if current_log_pattern.match(l):
        old_name=l+'_'+timesuffix
        print( f"found {l} to be renamed to {old_name}")
        os.rename( l, old_name)
        current_log_root=l
        old_logs=[]
    elif current_log_root and l.startswith(current_log_root):
        old_logs.append(l)
    else:
        if len(old_logs) > 0:
            if len(old_logs) > keep_count:
                old_to_delete.extend(old_logs[0:-keep_count])
            old_logs=[]
        else:
            print( f"skipping {l}")

if len(old_to_delete) == 0:
    print( "no old logs to delete")
else:
    for old_log in old_to_delete:
        print( f"removing {old_log}")
        os.unlink(old_log)



# This file is part of sarracenia.
# The sarracenia suite is Free and is proudly provided by the Government of Canada
# Copyright (C) Her Majesty The Queen in Right of Canada, Environment Canada, 2008-2020
#
# Sarracenia repository: https://github.com/MetPX/sarracenia
# Documentation: https://github.com/MetPX/sarracenia
#
"""
  plugins that use primarily the after_work entry point, normally 
  executed after the file transfer (either send or get) has completed.

  usually such plugins will contain a loop:

  for msg in worklist.ok
      do_something.

  to operate on all the files transferrred or processed successfully.

"""
pass



# This file is part of sarracenia.
# The sarracenia suite is Free and is proudly provided by the Government of Canada
# Copyright (C) Her Majesty The Queen in Right of Canada, Environment Canada, 2008-2020
#
# Sarracenia repository: https://github.com/MetPX/sarracenia
# Documentation: https://github.com/MetPX/sarracenia
#

import logging

logger = logging.getLogger(__name__)
"""
  This file was originally a place holder... thought this was not needed, then there was a stack trace.
  and setup.py gets leaves the directory out if not present.

  then started finding this convenient for common routines.

"""

pass



#!/usr/bin/python3
"""
  For Production deployments, see: part_check instead of this file (file_check)  

  As check re-reads the entire file, re-calculates from scratch, which is very inefficient 
  compared to part_check which takes care of checksum calc done as the files are downloaded.
  NOTE: it also deletes the downloaded file after checking!

Usage:

   callback work.check

STATUS:
  This is more for debugging, internal testing, and reference.
  20171212: not sure if it has been fixed after the local->new transition.

"""
import os, stat, time
from hashlib import md5
import logging

import sarracenia
from sarracenia.flowcb.work.check import Check

logger = logging.getLogger(__name__)

class Citypage_check(Check):

    def content_check(self, local_file) -> bool:
        logger.info(" start")
        f = open(local_file, 'rb')
        data = f.read()
        f.close()

        bad = False
        if "</siteData>" not in data.decode('iso8859-1'):
            logger.error( "does not have </siteData> in it, XML incomplete saving...")
            return False
        logger.info(" done")
        return True



# This file is part of sarracenia.
# The sarracenia suite is Free and is proudly provided by the Government of Canada
# Copyright (C) Her Majesty The Queen in Right of Canada, Environment Canada, 2008-2015
#
# Sarracenia repository: https://github.com/MetPX/sarracenia
# Documentation: https://github.com/MetPX/sarracenia
#
########################################################################
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; version 2 of the License.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
#
#

from sarracenia.transfer import Transfer

import sarracenia

import os, stat, subprocess, sys, time

import logging

logger = logging.getLogger(__name__)

#============================================================
# file protocol in sarracenia supports/uses :
#
# connect
# close
#
# if a source    : get    (remote,local)
#                  ls     ()
#                  cd     (dir)
#                  delete (path)
#
# require   logger
#           options.credentials
#           options.sendTo
#           options.batch
#           options.permDefault
#           options.permDirDefault
#     opt   options.byteRateMax
#     opt   options.bufsize


class File(Transfer):
    """
        Transfer sub-class for local file i/o.

    """
    def __init__(self, proto, options):
        super().__init__(proto, options)

        self.o.add_option("accelCpCommand", "str", "/usr/bin/cp %s %d")
        logger.debug("sr_file __init__")
        self.cwd = None

    def registered_as():
        return ['file']

    # cd
    def cd(self, path):
        """
           proto classes are used for remote sessions, so this 
           cd is for REMOTE directory... when file remote as a protocol it is for the source.
           should not change the "local" working directory when downloading.
        """
        logger.debug("sr_file cd %s" % path)
        #os.chdir(path)
        self.cwd = path
        self.path = path

    def check_is_connected(self):
        return True

    # chmod
    def chmod(self, perm, path):
        logger.debug("sr_file chmod %s %s" % ("{0:o}".format(perm), path))
        os.chmod(path, perm)

    # close
    def close(self):
        logger.debug("sr_file close")
        return

    # connect
    def connect(self):
        logger.debug("sr_file connect %s" % self.o.sendTo)

        self.recursive = True
        self.connected = True

        return True

    # delete
    def delete(self, path):
        p = os.path.join( self.cwd, path )
        logger.debug("sr_file rm %s" % p)
        os.unlink(p)

    # get
    def get(self,
            msg,
            remote_file,
            local_file,
            remote_offset=0,
            local_offset=0,
            length=0):

        remote_path = self.cwd + os.sep + remote_file

        logger.debug( "get %s %s (cwd: %s) %d" % (remote_path,local_file,os.getcwd(), local_offset))

        if not os.path.exists(remote_path):
            logger.warning("file to read not found %s" % (remote_path))
            return -1

        src = self.local_read_open(remote_path, remote_offset)
        dst = self.local_write_open(local_file, local_offset)

        # initialize sumalgo
        if self.sumalgo: self.sumalgo.set_path(remote_file)

        # download
        rw_length = self.read_write(src, dst, length)

        # close
        self.local_write_close(dst)

        return rw_length

    def getAccelerated(self, msg, remote_file, local_file, length=0):

        base_url = msg['baseUrl'].replace('file:', '')
        if base_url[-1] == '/':
            base_url = base_url[0:-1]
        arg1 = base_url + self.cwd + os.sep + remote_file
        arg1 = arg1.replace(' ', '\ ')
        arg2 = local_file

        cmd = self.o.accelCpCommand.replace('%s', arg1)
        cmd = cmd.replace('%d', arg2).split()

        logger.info("accel_cp:  %s" % ' '.join(cmd))
        p = subprocess.Popen(cmd)
        p.wait()
        if p.returncode != 0:
            return -1
        sz = os.stat(arg2).st_size
        return sz

    def getcwd(self):
        return self.cwd

    # ls
    def ls(self):
        logger.debug("sr_file ls")
        self.entries = {}
        self.root = self.path
        self.ls_python(self.path)
        return self.entries

    def ls_python(self, dpath):
        for x in os.listdir(dpath):
            dst = dpath + '/' + x
            if os.path.isdir(dst):
                if self.recursive: self.ls_python(dst)
                continue
            relpath = dst.replace(self.root, '', 1)
            if relpath[0] == '/': relpath = relpath[1:]

            self.entries[relpath] = sarracenia.stat(dst)


# file_insert
# called by file_process (general file:// processing)


def file_insert(options, msg):
    logger.debug("file_insert")

    fp = open(msg['relPath'], 'rb')
    if msg.partflg == 'i': fp.seek(msg['offset'], 0)

    ok = file_write_length(fp, msg, options.bufsize, msg.filesize, options)

    fp.close()

    return ok


def file_link(msg):

    try:
        os.unlink(msg['new_file'])
    except:
        pass
    try:
        os.link(msg['fileOp']['link'], os.path.join(self.cwd,msg['new_file']))
    except:
        return False

    msg.compute_local_checksum()
    msg.onfly_checksum = "{},{}".format(msg.sumflg, msg.local_checksum)

    return True


# file_process (general file:// processing)


def file_process(options):
    logger.debug("file_process")

    msg = options.msg

    # FIXME - MG - DOMINIC's LOCAL FILE MIRRORING BUG CASE
    # say file.txt does not exist
    # sequential commands in script
    # touch file.txt
    # mv file.txt newfile.txt
    # under libsrshim generate 3 amqp messages :
    # 1- download/copy file.txt
    # 2- move message 1 :  remove file.txt with newname newfile.txt
    # 3- move message 2 :  download newfile.txt with oldname file.txt
    # message (1) will never be processed fast enough ... and will fail
    # message (2) removing of a file not there is considered successfull
    # message (3) is the one that will guaranty the the newfile.txt is there and mirroring is ok.
    #
    # message (1) fails.. in previous version a bug was preventing an error (and causing file.txt rebirth with size 0)
    # In current version, returning that this message fails would put it under the retry process for ever and for nothing.
    # I decided for the moment to warn and to return success... it preserves old behavior without the 0 byte file generated

    if not os.path.isfile(msg['relPath']):
        logger.warning("%s moved or removed since announced" % msg['relPath'])
        return True

    try:
        curdir = self.cwd
    except:
        curdir = None

    if curdir != options.msg['new_dir']:
        os.chdir(options.msg['new_dir'])

    # try link if no inserts

    p=os.path.join(self.cwd,msg['relPath'])

    if msg.partflg == '1' or \
       (msg.partflg == 'p' and  msg.in_partfile) :
        ok = file_link(msg)
        if ok:
            if options.delete:
                try:
                    os.unlink(p)
                except:
                    logger.error("delete of link to %s failed" % p)
            return ok

    # This part is for 2 reasons : insert part
    # or copy file if preceeding link did not work
    try:
        ok = file_insert(options, msg)
        if options.delete:
            if msg.partflg.startswith('i'):
                logger.info("delete unimplemented for in-place part files %s" %
                            (msg['relPath']))
            else:
                try:
                    os.unlink(p)
                except:
                    logger.error("delete of %s after copy failed" % p)

        if ok: return ok

    except:
        logger.error('sr_file/file_process error')
        logger.debug('Exception details: ', exc_info=True)

    logger.error("could not copy %s in %s" % (p, msg['new_file']))

    return False


# file_write_length
# called by file_process->file_insert (general file:// processing)


def file_write_length(req, msg, bufsize, filesize, options):
    logger.debug("file_write_length")

    msg.onfly_checksum = None

    chk = msg.sumalgo
    logger.debug("file_write_length chk = %s" % chk)
    if chk: chk.set_path(msg['new_file'])

    # file should exists
    if not os.path.isfile(msg['new_file']):
        fp = open(msg['new_file'], 'w')
        fp.close()

    # file open read/modify binary
    fp = open(msg['new_file'], 'r+b')
    if msg.local_offset != 0: fp.seek(msg.local_offset, 0)

    nc = int(msg['length'] / bufsize)
    r = msg['length'] % bufsize

    # read/write bufsize "nc" times
    i = 0
    while i < nc:
        chunk = req.read(bufsize)
        fp.write(chunk)
        if chk: chk.update(chunk)
        i = i + 1

    # remaining
    if r > 0:
        chunk = req.read(r)
        fp.write(chunk)
        if chk: chk.update(chunk)

    if fp.tell() >= msg.filesize:
        fp.truncate()

    fp.close()

    h = options.msg.headers
    if options.permCopy and 'mode' in h:
        try:
            mod = int(h['mode'], base=8)
        except:
            mod = 0
        if mod > 0: os.chmod(msg['new_file'], mod)

    if options.timeCopy and 'mtime' in h and h['mtime']:
        os.utime(msg['new_file'],
                 times=(timestr2flt(h['atime']), timestr2flt(h['mtime'])))

    if chk:
        msg.onfly_checksum = "{},{}".format(chk.registered_as(), chk.value)

    return True


# file_truncate
# called under file_reassemble (itself and its file_insert_part)
# when inserting lastchunk, file may need to be truncated


def file_truncate(options, msg):

    # will do this when processing the last chunk
    # whenever that is

    if (not options.randomize) and (not msg.lastchunk): return

    try:
        lstat = sarracenia.stat(msg['target_file'])
        fsiz = lstat.st_size

        if fsiz > msg.filesize:
            fp = open(msg['target_file'], 'r+b')
            fp.truncate(msg.filesize)
            fp.close()

        msg['subtopic'] = msg['relPath'].split(os.sep)[1:-1]
        msg['_deleteOnPost'] |= set(['subtopic'])
        #msg.set_topic(options.post_topicPrefix,msg.target_relpath)

    except:
        pass



import json
import logging
import sarracenia
from sarracenia.postformat import PostFormat

logger = logging.getLogger(__name__)


class V03(PostFormat):
    """
       A class for controlling the format of messages are sent.
       internally All messages are represented as v03. 
       post format implement translations to other protocols for interop.

   """

    @staticmethod
    def content_type():
        return 'application/json'

    @staticmethod
    def mine(payload, headers, content_type, options) -> bool:
        """
          return true if the message is in this post format.
       """
        if content_type == V03.content_type():
            return True
        return False

    @staticmethod
    def importMine(body, headers, options) -> sarracenia.Message:
        """
          given a message in a wire format, with the given properties (or headers) in a dictionary,
          return the message as a normalized v03 message.
       """
        msg = sarracenia.Message()
        msg["_format"] = __name__.split('.')[-1].lower()


        try:
            msg.copyDict(json.loads(body))
        except Exception as ex:
            logger.warning('expected json, decode error: %s' % ex)
            logger.warning( f'body: {body}' )
            logger.debug('Exception details: ', exc_info=True)
            return None

        """ early sr3 versions used retPath, name changed by https://github.com/MetPX/sarracenia/issues/628
        """
        if 'retPath' in msg:
            msg['retrievePath'] = msg['retPath']
            del msg['retPath']

        if 'integrity' in msg:
            msg['identity'] = msg['integrity']
            del msg['integrity']

        """
              observed Sarracenia v2.20.08p1 and earlier have 'parts' header in v03 messages.
              bug, or implementation did not keep up. Applying Postel's Robustness principle: normalizing messages.
            """
        if ('parts' in msg
            ):  # bug in v2 code makes v03 messages with parts header.
            (m, s, c, r, n) = msg['parts'].split(',')
            if m == '1':
                msg['size'] = int(s)
            else:
                if m == 'i': m = 'inplace'
                elif m == 'p': m = 'partitioned'
                msg['blocks'] = {
                    'method': m,
                    'size': int(s),
                    'count': int(c),
                    'remainder': int(r),
                    'number': int(n)
                }

            del msg['parts']
        elif ('size' in msg):
            if (type(msg['size']) is str):
                msg['size'] = int(msg['size'])
        return msg

    @staticmethod
    def exportMine(body,options) -> (str, dict, str):
        """
           given a v03 (internal) message, produce an encoded version.
       """
        raw_body = json.dumps(body)

        headers = { 'topic': PostFormat.topicDerive(body,options) }

        return raw_body, headers, V03.content_type()



"""
   compresses the files and then writes their path to a SQLiteQueue. Requires the "persistqueue" module.
   usage:


    rxq_name <fname>
    flowCallback rxqueue_gzip.RxQueue_gzip

"""

import gzip
import logging
import os
import persistqueue
from sarracenia.flowcb import FlowCB

logger = logging.getLogger(__name__)


class RxQueue_gzip(FlowCB):

    def __init__(self,options):

        super().__init__(options,logger)
        self.o.add_option( option='rxq_name', kind='str' )

    def on_start(self):
        if not hasattr(self.o,'rxq_name'):
            logger.error("Missing rxq_name parameter")
            return
        self.rxq = persistqueue.SQLiteQueue(self.o.rxq_name, auto_commit=True)

    def after_work(self, worklist):
        for msg in worklist.ok:
            fname = f'{msg["new_dir"]}/{msg["new_file"]}'
            gzname = f'{fname}.gz'
            tname = f'{gzname}.tmp'
            if os.path.exists(fname):
                # Only try this if the uncompressed file actually exists
                gzip.open(tname, 'wb').write(open(fname,'rb').read())
                os.rename(tname, gzname)
                os.unlink(fname)
                self.rxq.put( gzname )
            else:
                self.rxq.put( fname )

        return None



"""

   An example of consuming data made available by a Sarracenia Data Pump.

"""
import sarracenia.moth
import sarracenia.moth.amqp
import sarracenia.credentials

import time
import socket

options = sarracenia.moth.default_options
options.update(sarracenia.moth.amqp.default_options)

options['broker'] = sarracenia.credentials.Credential(
    'amqps://anonymous:anonymous@hpfx.collab.science.gc.ca')

# binding tuple:  consists of prefix, exchange, rest.
# effect is to bind from queue using prefix/rest to exchange.
options['topicPrefix'] = ['v02', 'post']
options['bindings'] = [('xpublic', ['v02', 'post'], ['#'])]

# turn on debug output for these classes.
#options['settings'] = { 'sarracenia.moth.mqtt.MQTT': { 'logLevel':'debug' }}
#options['settings'] = { 'sarracenia.moth.amqp.AMQP': { 'logLevel':'debug' }}

# Note: queue name must start with q_<username> because server is configured to deny anything else.
#
options['queueName'] = 'q_anonymous_' + socket.getfqdn(
) + '_SomethingHelpfulToYou'

print('options: %s' % options)

h = sarracenia.moth.Moth.subFactory(options)

count = 0
while count < 5:
    m = h.getNewMessage()
    if m is not None:
        print("message: %s" % m)
        #content = m.getContent()
        #print("corresponding file: %s" % content)
        h.ack(m)
    time.sleep(0.1)
    count += 1

print(' got %d messages' % count)
h.cleanup()
h.close()
exit(count)



"""
Posts updated files of NOAA water level/temperature hydrometric data. Station site IDs provided
in the poll_noaa_stn_file. Compatible with Python 3.5+.

usage:
sample url: https://tidesandcurrents.noaa.gov/api/datagetter?range=1&station=9450460&product=water_temperature&units=metric&time_zone=gmt&application=web_services&format=csv

in an sr_poll configuration file::

	pollUrl http://tidesandcurrents.noaa.gov/api
        retrievePathPattern /datagetter?range=1&station={0:}&product={1:}&units=metric&time_zone=gmt&application=web_services&format=csv

	poll_noaa_stn_file [path/to/stn/file]
	callback noaa_hydrometric

sample station file::

        7|70678|9751639|Charlotte Amalie|US|VI|-4.0
        7|70614|9440083|Vancouver|US|WA|-8.0

The poll:
If poll_noaa_stn_file isn't set, it'll grab an up-to-date version of all station site code data from the 
NOAA website. The station list file is in the following format:
SourceID | SiteID | SiteCode | SiteName | CountryID | StateID | UTCOffset
Each station on its own line.
Posts the file on the exchange if the request returns a valid URL. 

in v2, one needed a matching downloader plugin, but in sr3 we can leverage the retrievePath feature
so that normalk downloader works, so only the poll one needed.

"""

import copy
import datetime
import logging

import os
import sarracenia
from sarracenia.flowcb import FlowCB
import urllib.request
import xml.etree.ElementTree as ET

logger = logging.getLogger(__name__)


class Noaa_hydrometric(FlowCB):
    def __init__(self, options):

        super().__init__(options,logger)

        # these options are only for the poll.
        self.o.add_option(option='poll_noaa_stn_file', kind='str')
        self.o.add_option( option='retrievePathPattern', kind='str', \
              default_value='datagetter?range=1&station={0:}&product={1:}&units=metric&time_zone=gmt&application=web_services&format=csv' )

        if self.o.identity_method.startswith('cod,'):
            m, v = self.o.identity_method.split(',')
            self.identity = {'method': m, 'value': v}

    def poll(self) -> list:

        # Make list of site codes to pass to http get request
        sitecodes = []
        if hasattr(self.o, 'poll_noaa_stn_file'):
            stn_file = self.o.poll_noaa_stn_file

            # Parse file to make list of all site codes
            try:
                with open(stn_file) as f:
                    for line in f:
                        items = line.split('|')
                        sitecodes.append(items[2])
                logger.info("poll_noaa used stn_file %s" % stn_file)

            except IOError as e:
                logger.error("poll_noaa couldn't open stn file: %s" % stn_file)

        else:
            # Grab station site codes from https://opendap.co-ops.nos.noaa.gov/stations/stationsXML.jsp
            tree = ET.parse(urllib.request.urlopen\
               ('https://opendap.co-ops.nos.noaa.gov/stations/stationsXML.jsp'))
            root = tree.getroot()
            for child in root:
                sitecodes.append(child.attrib['ID'])

        incoming_message_list = []
        # Every hour, form the link of water level/temp data to post
        for site in sitecodes:

            retrievePath = self.o.retrievePathPattern.format(site, 'water_temperature')
            url = self.o.pollUrl + retrievePath
            logger.info(f'polling {site}, polling: {url}')
            # Water temp request
            resp = urllib.request.urlopen(url).getcode()
            logger.info(f"poll_noaa file posted: {url} %s")
            mtime = datetime.datetime.utcnow().strftime('%Y%m%d_%H%M')

            fname = f'noaa_{mtime}_{site}_WT.csv'
            m = sarracenia.Message.fromFileInfo(fname, self.o)
            m['identity'] = self.identity
            m['retrievePath'] = retrievePath
            m['new_file'] = fname

            incoming_message_list.append(m)

            # Water level request
            retrievePath = self.o.retrievePathPattern.format(
                site, 'water_level') + '&datum=STND'
            url = self.o.pollUrl + retrievePath
            resp = urllib.request.urlopen(url).getcode()
            logger.info(f"poll_noaa file posted: {url}")

            fname = f'noaa_{mtime}_{site}_WL.csv'
            m = sarracenia.Message.fromFileInfo(fname, self.o)
            m['identity'] = self.identity
            m['retrievePath'] = retrievePath
            m['new_file'] = fname

            incoming_message_list.append(m)

        return incoming_message_list
